/***
 * Juliac
 * Copyright (C) 2018-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.api;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.util.NoSuchElementException;
import java.util.function.Predicate;
import java.util.logging.Logger;

import javax.annotation.processing.ProcessingEnvironment;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.juliac.api.generator.SourceCodeGeneratorItf;

/**
 * The interface of Juliac.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.7
 */
public interface JuliacItf {

	public void close();

	/**
	 * Generate the source code of a factory class for the specified ADL.
	 *
	 * @param adl         the fully-qualified name of the ADL descriptor
	 * @param targetname  the fully-qualified name of the factory class
	 * @throws IllegalArgumentException
	 *                 if none of the registered ADL parsers accept the
	 *                 specified ADL descriptor
	 */
	void generate(String adl, String targetname) throws IOException;

	/**
	 * Generate the source code for the specified membrane descriptor.
	 *
	 * @param ctrlDesc  the membrane descriptor
	 * @throws IllegalArgumentException  if no source code generator is suitable
	 */
	void generate(String ctrlDesc) throws IOException, InstantiationException;

	/**
	 * Generate the source code for the specified membrane descriptor defined in
	 * the specified source code element.
	 *
	 * @param ctrlDesc  the membrane descriptor
	 * @throws IllegalArgumentException  if no source code generator is suitable
	 */
	void generate(String ctrlDesc, Object source)
	throws IOException, InstantiationException;

	/**
	 * Generate the source code for the specified component.
	 *
	 * @param type            the component type
	 * @param controllerDesc  the controller descriptor
	 * @param contentDesc     the content descriptor
	 * @param source          the code element for this component definition
	 * @return  the initializer class generator
	 * @throws IllegalArgumentException  if no source code generator is suitable
	 */
	SourceCodeGeneratorItf generate(Type type, Object controllerDesc,
			Object contentDesc, Object source) throws IOException;

	/**
	 * Generate the source code for the specified component descriptor.
	 *
	 * The descriptor is expected to be of the form
	 * <code>[name/signature/bool/bool/bool;...,controllerDesc,contentDesc]</code>.
	 *
	 * @param desc  the component descriptor
	 * @throws IllegalArgumentException
	 *                 if the descriptor does not match the expected syntax or
	 *                 if no source code generator is suitable
	 */
	void generateForDesc(String desc)
	throws IOException, InstantiationException;

	/**
	 * Dump the code generated by the specified source code generator in its
	 * corresponding file. Do not perform the generation if the type has already
	 * been generated or can be loaded from in the classpath.
	 */
	void generateSourceCode(SourceCodeGeneratorItf scg) throws IOException;

	/**
	 * Dump the code generated by the specified source code generator in its
	 * corresponding file. Return the URI of the file.
	 */
	URI generateSourceCodeOverride(SourceCodeGeneratorItf scg)
			throws IOException;

	/**
	 * Return <code>true</code> if the specified type has already been generated
	 * or exists in the classpath.
	 */
	boolean hasBeenGenerated(String name);

	/**
	 * Compile the input and generated source code.
	 */
	void build() throws IOException;

	/**
	 * Add a generated file.
	 *
	 * @param sf          the source file containing the generated code
	 * @param sourceType  the source type from which the code has been generated
	 */
	void addGenerated(SourceFileItf sf, Object sourceType);

	/**
	 * Return <code>true</code> if the specified qualified name
	 * corresponds to a generated type name.
	 */
	boolean contains(String qname);

	/**
	 * Return the source type corresponding to the specified qualified
	 * name.
	 */
	Object getSourceType(String qname);

	/**
	 * Load the requested class. If the specified name corresponds to a
	 * primitive type, return the corresponding class.
	 *
	 * @param name  the fully qualified name of the requested class
	 * @return      the corresponding class
	 * @throws RuntimeClassNotFoundException  if the class cannot be loaded
	 */
	<T> Class<T> loadClass(String name);

	/**
	 * Instantiate the specified class by calling the default constructor.
	 * Wrap exceptions thrown by {@link Class#newInstance()} ({@link
	 * InstantiationException} and {@link IllegalAccessException}) into {@link
	 * JuliacRuntimeException}.
	 *
	 * @param <T>        the type of the class to instantiate
	 * @param className  the name of the class to instantiate
	 * @return           the instance
	 * @throws JuliacRuntimeException  in case of error during the instantiation
	 */
	<T> T instantiate(Class<T> cl);

	/**
	 * Unregister all services.
	 */
	void unregisterAll();

	/**
	 * Register the specified service implementation.
	 *
	 * @param itf      the service type
	 * @param impl  the service implementation
	 */
	<T extends JuliacModuleItf> void register(Class<T> itf, T impl);

	/**
	 * Retrieve the services implementing the specified service type.
	 *
	 * @param itf  the service type
	 * @return     the registered service implementations
	 */
	<T extends JuliacModuleItf> Iterable<T> lookup(Class<T> itf);

	/**
	 * Retrieve the services implementing the specified service type and
	 * accepted by the specified filter and value.
	 *
	 * @param itf    the service type
	 * @param value  the value for evaluating the predicate
	 * @return       the first available service implementation
	 * @throws IllegalArgumentException  if no service is registered
	 */
	public <T extends JuliacModuleItf & Predicate<P>,P> T lookupAndFilter(
		Class<T> itf, P value );

	/**
	 * Retrieve the first available service implementing the specified
	 * service types.
	 *
	 * @param itf  the service type
	 * @return     the first available registered service implementation
	 * @throws NoSuchElementException  if no service is registered
	 */
	<T extends JuliacModuleItf> T lookupFirst(Class<T> itf);

	/**
	 * Unregister the specified service implementation.
	 *
	 * @param itf   the service type
	 * @param impl  the service implementation
	 */
	<T extends JuliacModuleItf> void unregister(Class<T> itf, T impl);

	void setProcessingEnvironment(ProcessingEnvironment pe);

	ProcessingEnvironment getProcessingEnvironment();

	void setJuliaCfgFiles(String juliaConfig);

	String getJuliaCfgFiles();

	void setPkgRoot(String pkgRoot);

	String getPkgRoot();

	void setBaseDir(File baseDir);

	File getBaseDir();

	String[] getSrcs();

	/**
	 * Add the specified location to the list of locations. A location
	 * denotes either an absolute path or a path relative to
	 * {@link #baseDir}, for a directory or a jar file. All source files
	 * contained in the specified location are registered for being
	 * compiled when requested (see {@link #build()}) or processed by
	 * the juliac-spoon module.
	 *
	 * @param src  the location to add
	 * @throws IOException
	 *      if the location does not exist or an error occurred while fetching
	 *      the source files contained in the location
	 */
	void addSrc(String src) throws IOException;

	/**
	 * Invoke {@link #addSrc(String)} for each element contained in the
	 * specified array of locations.
	 */
	void addSrcs(String[] srcs) throws IOException;

	/**
	 * Add the specified resource file to the class directory.
	 *
	 * @param filename  the name of the resource file
	 * @param sb        the content of the resource file
	 */
	void addResource(String filename, StringBuilder sb) throws IOException;

	String[] getSrclibs();

	/**
	 * Add the specified location to the list of library locations. A location
	 * denotes either an absolute path or a path relative to {@link #baseDir},
	 * for a directory or a jar file.
	 *
	 * @param src  the location to add
	 */
	void addSrclib(String srclib);

	/**
	 * Invoke {@link #addSrclibs(String)} for each element contained in the
	 * specified array of locations.
	 */
	void addSrclibs(String[] srclibs);

	File getGenDir() throws IOException;

	String getGenDirName();

	void setGenDirName(String genDirName) throws IOException;

	File getClassDir() throws IOException;

	String getClassDirName();

	void setClassDirName(String classDirName) throws IOException;

	ClassLoader getClassLoader();

	void setClassLoader(ClassLoader classLoader);

	/**
	 * Return the default logger.
	 */
	Logger getLogger();

	/**
	 * Return the Java release version number to be used when compiling code.
	 * Used in the juliac-jcapi and juliac-spoon-core (because Spoon relies on
	 * JDT that needs this piece of information) modules.
	 *
	 * @since 2.8
	 */
	String getJavaReleaseVersionNumber();
}
