/***
 * Juliac
 * Copyright (C) 2017-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.adlet.core.proxy;

import java.lang.reflect.Modifier;
import java.util.List;

import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.adlet.api.generator.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.adlet.core.helper.TypeElementHelper;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.core.helper.JuliacHelper;
import org.objectweb.fractal.juliac.core.opt.ClassGenerator;

import com.google.common.base.Joiner;

/**
 * Root class for {@link InterfaceType}-based proxy class generators.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.7
 */
public abstract class AbstractProxyClassGenerator extends ClassGenerator
implements ProxyClassGeneratorItf {

	protected InterfaceType it;
	protected TypeElement proxycl;
	protected String pkgRoot;
	protected MembraneDesc<?> membraneDesc;
	protected boolean mergeable;

	public AbstractProxyClassGenerator() {}

	public AbstractProxyClassGenerator(
		InterfaceType it, TypeElement proxycl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		setInterfaceType(it,proxycl,pkgRoot);
		setMembraneDesc(membraneDesc);
		setMergeable(mergeable);
	}

	public void setInterfaceType( InterfaceType it, Object o, String pkgRoot ) {

		/*
		 * o is null when the corresponding signature of the interface type
		 * cannot be retrieved as a TypeElement. This happens when requesting
		 * the generation of Fractal interface implementations for attribute
		 * implementation classes. The generation has been previously done with
		 * AttributeControlFcInterfaceImplementationClassGenerator, and proxycl
		 * can stay null since the only method that will be called is
		 * getTargetTypeName() that does not use this parameter. This method is
		 * called to check that the class name is present in the compilation
		 * round.
		 */

		if( o!=null && !(o instanceof TypeElement) ) {
			final String msg =
				"Expected type TypeElement for parameter "+o+" signature "+
				it.getFcItfSignature();
			throw new UnsupportedOperationException(msg);
		}
		TypeElement proxycl = (TypeElement) o;

		/*
		 * Check that the specified interface type corresponds to a Java
		 * interface (not a class.)
		 *
		 * Generating proxies based on class types leads to inconsistencies.
		 * Indeed, when retrieving public methods, the methods from
		 * java.lang.Object are returned. First, generating proxies for these
		 * methods may seem counterintuitive. Second, and most importantly, some
		 * of these methods (e.g. getClass) are final: trying to override them
		 * in the proxy class leads to compilation errors.
		 *
		 * This problem was raised by Philippe, Romain and Daniel in a case
		 * where FraSCAti was erroneously trying to generate code for a service
		 * interface of type Void.
		 */
		if( proxycl!=null && ! proxycl.getKind().isInterface() ) {
			final String msg =
				"Signature of InterfaceType "+it.getFcItfSignature()+
				" must correspond to a Java interface (not a class)";
			throw new IllegalArgumentException(msg);
		}

		this.it = it;
		this.proxycl = proxycl;
		this.pkgRoot = pkgRoot;
	}

	public void setMembraneDesc( MembraneDesc<?> membraneDesc ) { this.membraneDesc = membraneDesc; }
	public void setMergeable( boolean mergeable ) { this.mergeable = mergeable; }


	/**
	 * Return the name of the class generated by this generator.
	 */
	public String getTargetTypeName() {
		String javaItfSignature = it.getFcItfSignature();
		String name = JuliacHelper.getJuliacGeneratedStrongTypeName(javaItfSignature);
		String suffix = getClassNameSuffix();
		String targetClassName = pkgRoot + name + suffix;
		return targetClassName;
	}

	@Override
	public int getTypeModifiers() {
		int modifiers = mergeable ? Modifier.ABSTRACT : super.getTypeModifiers();
		return modifiers;
	}

	/**
	 * Generate the source code of the proxy class.
	 *
	 * @param cv  the visitor where the source code is generated
	 */
	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {

		List<? extends Element> elements =
				TypeElementHelper.getAllMembers(proxycl);

//        Method[] proxymethods =
//            ClassHelper.getMethodsWithMostSpecializedReturnType(proxycl);

		for( Element element : elements ) {
			if( element.getKind().equals(ElementKind.METHOD) ) {
				generateProxyMethod(cv,(ExecutableElement)element);
			}
		}
	}

	@Override
	public String[] getGenericTypeParameters() {
		List<? extends TypeParameterElement> tps = proxycl.getTypeParameters();
		String[] typeParameters = new String[tps.size()];
		for (int i = 0; i < typeParameters.length; i++) {
			typeParameters[i] = tps.get(i).toString();
		}
//        GenericClass proxycl = GenericClass.get(this.proxycl);
//        String[] tps = proxycl.getTypeParameters();
		return typeParameters;
	}

	public void generateProxyMethod(
		ClassSourceCodeVisitor cv, ExecutableElement proxym ) {

		int modifiers = Modifier.PUBLIC;

		List<? extends TypeParameterElement> tps = proxym.getTypeParameters();
		String[] typeParameters = new String[tps.size()];
		for (int i = 0; i < typeParameters.length; i++) {
			typeParameters[i] = tps.get(i).toString();
		}

		String returnType = proxym.getReturnType().toString();
		String name = proxym.getSimpleName().toString();

		List<? extends VariableElement> ps = proxym.getParameters();
		String[] parameters = new String[ps.size()];
		for (int i = 0; i < parameters.length; i++) {
			parameters[i] = ps.get(i).asType().toString()+" arg"+i;
		}

		List<? extends TypeMirror> thrownTypes = proxym.getThrownTypes();
		String[] exceptions = new String[thrownTypes.size()];
		for (int i = 0; i < exceptions.length; i++) {
			exceptions[i] = thrownTypes.get(i).toString();
		}

//        GenericClass gc = GenericClass.get(proxycl);
//        String[] typeParameters = gc.getTypeParameters(proxym);
//        String returnType = gc.getGenericReturnType(proxym);
//        String name = proxym.getName();
//        String[] parameters = gc.getParameters(proxym);
//        String[] exceptions = gc.getGenericExceptionTypes(proxym);

		/*
		 * Declare all method parameters as final.
		 * Needed when the parameter is used in an anonymous inner class.
		 * Used by Tinfi for oneway methods (Thread created.)
		 */
		for (int i = 0; i < parameters.length; i++) {
			if( ! parameters[i].startsWith("final") ) {
				parameters[i] = "final "+parameters[i];
			}
		}

		if(mergeable) {
			// _super_ method declaration to propagate the invocation to preceding layers
			int mergeablemodifiers = Modifier.PUBLIC + Modifier.ABSTRACT;
			String mergeablename = "_super_"+name;
			BlockSourceCodeVisitor mv =
				cv.visitMethod(
					mergeablemodifiers, typeParameters, returnType,
					mergeablename, parameters, exceptions );
			mv.visitEnd();
		}

		// Method declaration
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				modifiers, typeParameters, returnType,
				name, parameters, exceptions );

		// Method body
		generateProxyMethodBody(mv,proxym);
		mv.visitEnd();
	}

	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, ExecutableElement proxym ) {

		// Call to the delegate
		generateProxyMethodBodyDelegatingCode(mv,proxym);

		// Return the result
		generateProxyMethodBodyReturn(mv,proxym);
	}

	public void generateProxyMethodBodyDelegatingCode(
		BlockSourceCodeVisitor mv, ExecutableElement proxym ) {

		List<? extends VariableElement> parameters = proxym.getParameters();
		TypeMirror[] ptypes = new TypeMirror[parameters.size()];
		for (int i = 0; i < ptypes.length; i++) {
			ptypes[i] = parameters.get(i).asType();
		}

		String rtypename = proxym.getReturnType().toString();

		List<? extends TypeParameterElement> tps = proxym.getTypeParameters();
		String[] tpnames = new String[tps.size()];
		for (int i = 0; i < tpnames.length; i++) {
			tpnames[i] = tps.get(i).getSimpleName().toString();
		}

//        Class<?>[] ptypes = proxym.getParameterTypes();
//        GenericClass gc = GenericClass.get(proxycl);
//        String rtypename = gc.getGenericReturnType(proxym);
//        String[] tpnames = MethodHelper.getTypeParameterNames(proxym);

		String line = "";
		if( ! rtypename.equals("void") ) {
		   line += rtypename+" ret = ";
		}

		String delegatingInstance = getDelegatingInstance(proxym);
		String delegatingMethodName = getDelegatingMethodName(proxym);
		line+= delegatingInstance+".";
		if( tpnames.length > 0 ) {
			line+="<"+Joiner.on(",").join(tpnames)+">";
		}
		line+=delegatingMethodName+"(";
		for (int j = 0; j < ptypes.length; j++) {
			if(j!=0)  line+=",";
			line+="arg"+Integer.toString(j);
		}
		mv.visitIns(line+")");
	}

	public String getDelegatingMethodName( ExecutableElement proxym ) {
		String name = proxym.getSimpleName().toString();
		if(mergeable) {
			return "_super_" + name;
		}
		return name;
	}

	@Override
	public void generateProxyMethodBodyReturn( BlockSourceCodeVisitor mv, ExecutableElement proxym ) {
		String rtypename = proxym.getReturnType().toString();
//        GenericClass gc = GenericClass.get(proxycl);
//        String rtypename = gc.getGenericReturnType(proxym);
		if( ! rtypename.equals("void") ) {
			mv.visitIns("return","ret");
		}
	}

	@Override
	public TypeElement getSourceType() {
		return proxycl;
	}
}
