/***
 * Juliac
 * Copyright (C) 2017-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.adlet.core.proxy;

import java.lang.reflect.Modifier;

import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.TypeElement;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.BasicComponentInterface;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.core.visit.CodeHelper;

/**
 * This class generates the source code of Fractal interface implementations.
 *
 * This class is specialized for the annotation processing tool of the JDK.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.7
 */
public class InterfaceImplementationClassGenerator
extends AbstractProxyClassGenerator {

	public InterfaceImplementationClassGenerator(
		InterfaceType it, TypeElement proxycl, String pkgRoot,
		boolean mergeable ) {

		super(it,proxycl,pkgRoot,null,mergeable);
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source class generator.
	 */
	public boolean match() {
		return true;
	}

	/**
	 * Return the name of the suffix that is appended to generated class names.
	 */
	public String getClassNameSuffix() {
		return "FcItf";
	}

	/**
	 * @throws IllegalArgumentException  if mergeable is <code>true</code>
	 */
	@Override
	public void setMergeable(boolean mergeable) {
		if (mergeable) {
			final String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	// --------------------------------------------------------------------
	// ProxyClassGenerator methods
	// --------------------------------------------------------------------

	@Override
	public String getSuperClassName() {
		return BasicComponentInterface.class.getName();
	}

	@Override
	public String[] getImplementedInterfaceNames() {
		String signature = it.getFcItfSignature();
//        String signature =
//            InterfaceTypeHelper.getInterfaceTypeSignature(it,proxycl);
		return new String[] { signature };
	}

	@Override
	public void generateFields( ClassSourceCodeVisitor cv ) {
		generateFieldImpl(cv);
	}

	@Override
	public void generateConstructors( ClassSourceCodeVisitor cv ) {

		// public constructor()
		cv.visitConstructor(Modifier.PUBLIC, null, null, null).visitEnd();

		// public constructor(Component,String,Type,boolean,Object)
		String[] parameters = new String[] {
				Component.class.getName() + " component", "String s",
				Type.class.getName() + " type", "boolean flag", "Object obj" };
		BlockSourceCodeVisitor mv = cv.visitConstructor(Modifier.PUBLIC, null,
				parameters, null);
		mv.visitIns("super(component,s,type,flag,obj)").visitEnd();
	}

	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {
		super.generateMethods(cv);
		generateMethodGetFcItfImpl(cv);
		generateMethodSetFcItfImpl(cv);
	}

	@Override
	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, ExecutableElement proxym ) {

		BlockSourceCodeVisitor then = mv.visitIf("impl","==","null");
		then.visitIns(CodeHelper._throw(NullPointerException.class,
				"\"Trying to invoke a method on a client or server interface "+
				"whose complementary interface is not bound.\""));
		then.visitEnd();

		super.generateProxyMethodBody(mv,proxym);
	}

	public String getDelegatingInstance( ExecutableElement proxym ) {
		return "impl";
	}


	// -------------------------------------------------------------------
	// Implementation specific
	// -------------------------------------------------------------------

	private void generateFieldImpl( ClassSourceCodeVisitor cv ) {
		String signature = it.getFcItfSignature();
//        String signature =
//            InterfaceTypeHelper.getInterfaceTypeSignature(it,proxycl);
		cv.visitField(Modifier.PRIVATE, signature, "impl", null);
	}

	private void generateMethodGetFcItfImpl( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv = cv.visitMethod(Modifier.PUBLIC, null,
				"Object", "getFcItfImpl", null, null);
		mv.visitIns("return","impl").visitEnd();
	}

	private void generateMethodSetFcItfImpl( ClassSourceCodeVisitor cv ) {
		String signature = it.getFcItfSignature();
		BlockSourceCodeVisitor mv = cv.visitMethod(Modifier.PUBLIC, null,
				"void", "setFcItfImpl", new String[] { "Object obj" }, null);
		mv.visitSet("impl","("+signature+")obj").visitEnd();
	}
}
