/***
 * Juliac
 * Copyright (C) 2018-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.adlet.opt.comp;

import java.lang.reflect.Method;
import java.util.List;
import java.util.Set;

import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.TypeElement;
import javax.lang.model.element.VariableElement;
import javax.lang.model.util.ElementFilter;
import javax.lang.model.util.Elements;

import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.fraclet.annotations.Lifecycle;
import org.objectweb.fractal.fraclet.types.Step;
import org.objectweb.fractal.juliac.adlet.core.helper.TypeElementHelper;
import org.objectweb.fractal.juliac.api.JuliacItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.core.opt.FCSourceCodeGeneratorItf;

/**
 * Initializer generator for Fractal components with a component-based control
 * membrane.
 *
 * This class is specialized for the annotation processing tool of the JDK.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.8
 */
public class InitializerCompCtrlClassGenerator
extends org.objectweb.fractal.juliac.fraclet.opt.comp.InitializerCompCtrlClassGenerator {

	public InitializerCompCtrlClassGenerator(
		JuliacItf jc, FCSourceCodeGeneratorItf fcscg,
		MembraneDesc<?> membraneDesc, ComponentType ct,
		Object contentDesc, Object source ) {

		super(jc,fcscg,membraneDesc,ct,contentDesc,source);
	}

	@Override
	protected void generateLifeCycleCreate(
		BlockSourceCodeVisitor mv, String contentClassName ) {

		TypeElement cl = getTypeElement(contentClassName);
		List<? extends Element> members = TypeElementHelper.getAllMembers(cl);
		List<ExecutableElement> methods = ElementFilter.methodsIn(members);

//        methods = MethodHelper.removeOverridden(methods);

		boolean decl = false;
		for (ExecutableElement method : methods) {
			Lifecycle lifecycle = method.getAnnotation(Lifecycle.class);
			if( lifecycle != null ) {
				List<? extends VariableElement> types = method.getParameters();
				if( types.size() != 0 ) {
					final String msg =
						"Lifecycle callback method "+method+
						" cannot declare parameters";
					throw new JuliacRuntimeException(msg);
				}
				Step step = lifecycle.step();
				if( step.equals(Step.CREATE) ) {
					Set<Modifier> mods = method.getModifiers();
					if( mods.contains(Modifier.PUBLIC) ) {
						mv.visitIns(
							"(("+contentClassName+")content)."+
							method.getSimpleName()+"()");
					}
					else {
						// Use reflection to invoke non public methods
						if(!decl) {
							mv.visitVar(Method.class.getName(),"m","null");
							decl = true;
						}
						TypeElement enclosingElement = (TypeElement)
							method.getEnclosingElement();
						mv.visitSet(
							"m",
							enclosingElement.getQualifiedName().toString()+
							".class.getDeclaredMethod(\""+
							method.getSimpleName()+"\")");
						mv.visitIns("m.setAccessible(true)");
						mv.visitIns("m.invoke(content)");
					}
				}
			}
		}
	}

	@Override
	protected Object loadClass( String signature ) {

		ProcessingEnvironment env = jc.getProcessingEnvironment();

		Elements elements = env.getElementUtils();
		TypeElement cl = elements.getTypeElement(signature);
		return cl;
	}

	private TypeElement getTypeElement( String contentClassName ) {

		ProcessingEnvironment env = jc.getProcessingEnvironment();

		Elements elements = env.getElementUtils();
		TypeElement cl = elements.getTypeElement(contentClassName);

		if( cl == null ) {
			/*
			 * cl is null when the type has been generated and cannot thus be
			 * retrieved as a TypeElement. In this case rely on the source type
			 * that triggered the generation.
			 *
			 * This is the case for AttributesImpl suffixed classes generated by
			 * org.objectweb.fractal.juliac.adlet.proxy.AttributeControllerGenerator
			 * The generated AttributesImpl suffixed classes extend their source
			 * type and do not contain lifecycle related code that may be used
			 * by this method. This is then safe to retrieve the corresponding
			 * information contained in the source type.
			 */
			boolean b = jc.contains(contentClassName);
			if(b) {
				cl = (TypeElement) jc.getSourceType(contentClassName);
			}
			else {
				final String msg =
					"No registered source type for "+contentClassName;
				throw new JuliacRuntimeException(msg);
			}
		}

		return cl;
	}
}
