/***
 * Juliac
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Jonathan Labejof
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.core.proxy;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.generator.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.lang.ClassHelper;
import org.objectweb.fractal.juliac.commons.lang.GenericClass;
import org.objectweb.fractal.juliac.commons.lang.reflect.MethodHelper;
import org.objectweb.fractal.juliac.core.helper.JuliacHelper;
import org.objectweb.fractal.juliac.core.opt.ClassGenerator;

import com.google.common.base.Joiner;

/**
 * Root class for {@link InterfaceType}-based proxy class generators.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @author Jonathan Labejof <Jonathan.Labejof@inria.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille.fr>
 */
public abstract class AbstractProxyClassGenerator extends ClassGenerator
implements ProxyClassGeneratorItf {

	protected InterfaceType it;
	protected Class<?> proxycl;
	protected GenericClass proxygcl;
	protected String pkgRoot;
	protected MembraneDesc<?> membraneDesc;
	protected boolean mergeable;

	public AbstractProxyClassGenerator() {}

	public AbstractProxyClassGenerator(
		InterfaceType it, Class<?> proxycl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		setInterfaceType(it,proxycl,pkgRoot);
		setMembraneDesc(membraneDesc);
		setMergeable(mergeable);
	}

	@Override
	public void setInterfaceType( InterfaceType it, Object o, String pkgRoot ) {

		if( ! (o instanceof Class<?>) ) {
			final String msg = "Expected type Class<?> for parameter "+o;
			throw new UnsupportedOperationException(msg);
		}
		Class<?> proxycl = (Class<?>) o;

		/*
		 * Check that the specified interface type corresponds to a Java
		 * interface (not a class.)
		 *
		 * Generating proxies based on class types leads to inconsistencies.
		 * Indeed, when retrieving public methods, the methods from
		 * java.lang.Object are returned. First, generating proxies for these
		 * methods may seem counterintuitive. Second, and most importantly, some
		 * of these methods (e.g. getClass) are final: trying to override them
		 * in the proxy class leads to compilation errors.
		 *
		 * This problem was raised by Philippe, Romain and Daniel in a case
		 * where FraSCAti was erroneously trying to generate code for a service
		 * interface of type Void.
		 */
		if( ! proxycl.isInterface() ) {
			final String msg =
				"Signature of InterfaceType "+it.getFcItfSignature()+
				" must correspond to a Java interface (not a class)";
			throw new IllegalArgumentException(msg);
		}

		this.it = it;
		this.proxycl = proxycl;
		this.proxygcl = new GenericClass(proxycl);
		this.pkgRoot = pkgRoot;
	}

	@Override
	public void setMembraneDesc( MembraneDesc<?> membraneDesc ) { this.membraneDesc = membraneDesc; }

	@Override
	public void setMergeable( boolean mergeable ) { this.mergeable = mergeable; }


	/**
	 * Return the name of the class generated by this generator.
	 */
	@Override
	public String getTargetTypeName() {
		String javaItfSignature = it.getFcItfSignature();
		String name = JuliacHelper.getJuliacGeneratedStrongTypeName(javaItfSignature);
		String suffix = getClassNameSuffix();
		String targetClassName = pkgRoot + name + suffix;
		return targetClassName;
	}

	@Override
	public int getTypeModifiers() {
		int modifiers = mergeable ? Modifier.ABSTRACT : super.getTypeModifiers();
		return modifiers;
	}

	/**
	 * Generate the source code of the proxy class.
	 *
	 * @param cv  the visitor where the source code is generated
	 */
	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {

		super.generateMethods(cv);

		Method[] proxymethods =
			ClassHelper.getMethodsWithMostSpecializedReturnType(proxycl);

		for( Method proxym : proxymethods ) {
			generateProxyMethod(cv,proxym);
		}
	}

	@Override
	public String[] getGenericTypeParameters() {
		String[] tps = proxygcl.getTypeParameters();
		return tps;
	}

	@Override
	public void generateProxyMethod(
		ClassSourceCodeVisitor cv, Method proxym ) {

		int modifiers = Modifier.PUBLIC;
		String[] typeParameters = proxygcl.getTypeParameters(proxym);
		String returnType = proxygcl.getGenericReturnType(proxym);
		String name = proxym.getName();
		String[] parameters = proxygcl.getParameters(proxym);
		String[] exceptions = proxygcl.getGenericExceptionTypes(proxym);

		/*
		 * Declare all method parameters as final.
		 * Needed when the parameter is used in an anonymous inner class.
		 * Used by Tinfi for oneway methods (Thread created.)
		 */
		for (int i = 0; i < parameters.length; i++) {
			if( ! parameters[i].startsWith("final") ) {
				parameters[i] = "final "+parameters[i];
			}
		}

		if(mergeable) {
			// _super_ method declaration to propagate the invocation to preceding layers
			int mergeablemodifiers = Modifier.PUBLIC + Modifier.ABSTRACT;
			String mergeablename = "_super_"+name;
			BlockSourceCodeVisitor mv =
				cv.visitMethod(
					mergeablemodifiers, typeParameters, returnType,
					mergeablename, parameters, exceptions );
			mv.visitEnd();
		}

		// Method declaration
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				modifiers, typeParameters, returnType,
				name, parameters, exceptions );

		// Method body
		generateProxyMethodBody(mv,proxym);
		mv.visitEnd();
	}

	@Override
	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, Method proxym ) {

		// Call to the delegate
		generateProxyMethodBodyDelegatingCode(mv,proxym);

		// Return the result
		generateProxyMethodBodyReturn(mv,proxym);
	}

	@Override
	public void generateProxyMethodBodyDelegatingCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		Class<?>[] ptypes = proxym.getParameterTypes();
		String rtypename = proxygcl.getGenericReturnType(proxym);
		String[] tpnames = MethodHelper.getTypeParameterNames(proxym);

		String line = "";
		if( ! rtypename.equals("void") ) {
		   line += rtypename+" ret = ";
		}

		String delegatingInstance = getDelegatingInstance(proxym);
		String delegatingMethodName = getDelegatingMethodName(proxym);
		line+= delegatingInstance+".";
		if( tpnames.length > 0 ) {
			line+="<"+Joiner.on(",").join(tpnames)+">";
		}
		line+=delegatingMethodName+"(";
		for (int j = 0; j < ptypes.length; j++) {
			if(j!=0)  line+=",";
			line+="arg"+Integer.toString(j);
		}
		mv.visitIns(line+")");
	}

	@Override
	public String getDelegatingMethodName( Method proxym ) {
		String name = proxym.getName();
		if(mergeable) {
			return "_super_" + name;
		}
		return name;
	}

	@Override
	public void generateProxyMethodBodyReturn(
		BlockSourceCodeVisitor mv, Method proxym ) {

		String rtypename = proxygcl.getGenericReturnType(proxym);
		if( ! rtypename.equals("void") ) {
			mv.visitIns("return","ret");
		}
	}

	/**
	 * @since 2.8
	 */
	@Override
	public Class<?> getSourceType() {
		return proxycl;
	}
}
