/***
 * Juliac
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.core.proxy;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;

import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.control.attribute.CloneableAttributeController;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;

/**
 * This class generates attribute controller implementations.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille.fr>
 */
public class AttributeControllerClassGenerator
extends AbstractProxyClassGenerator {

	public AttributeControllerClassGenerator(
		InterfaceType it, Class<?> proxycl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,proxycl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source code generator.
	 */
	@Override
	public boolean match() {
		return true;
	}

	/**
	 * Return the name of the suffix that is appended to generated class names.
	 */
	@Override
	public String getClassNameSuffix() {
		return "AC";
	}

	@Override
	public void setMergeable( boolean mergeable ) {
		if(mergeable) {
			String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	@Override
	public String[] getImplementedInterfaceNames() {
		return new String[]{
			it.getFcItfSignature(),
			CloneableAttributeController.class.getName()};
	}

	@Override
	public void generateFields( ClassSourceCodeVisitor cv ) {

		/*
		 * Generate fields.
		 *
		 * One field per attribute.
		 * A boolean _ prefixed field per settable attribute.
		 * This field is set to true when the field has been modified.
		 */
		List<String> attrs = new ArrayList<>();
		settableAttrs = new ArrayList<>();
		Method[] proxymethods = proxycl.getMethods();
		for (int i = 0; i < proxymethods.length; i++) {

			Method m = proxymethods[i];
			String mname = m.getName();
			if( ! (mname.startsWith("get") || mname.startsWith("set")) ) {
				continue;
			}

			String attrname = mname.substring(3);
			if( mname.startsWith("set") ) {
				settableAttrs.add(attrname);
				if( ! attrs.contains(attrname) ) {
					Class<?> type = m.getParameterTypes()[0];
					cv.visitField(Modifier.PRIVATE, type.getName(), attrname, null);
				}
				cv.visitField(Modifier.PRIVATE, "boolean", "_"+attrname, null);
			}
			else {  // mname.startsWith("get")
				if( ! attrs.contains(attrname) ) {
					Class<?> type = m.getReturnType();
					cv.visitField(Modifier.PRIVATE, type.getName(), attrname, null);
				}
			}

			attrs.add(attrname);
		}
	}
	private List<String> settableAttrs;

	@Override
	public void generateConstructors( ClassSourceCodeVisitor cv ) {
		// public constructor()
		BlockSourceCodeVisitor mv =
			cv.visitConstructor(Modifier.PUBLIC,null,null,null);
		mv.visitEnd();
	}

	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {
		super.generateMethods(cv);
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "void", "cloneFcAttributes",
				new String[]{
					AttributeController.class.getName()+" attributecontroller"
				},
				null );
		for (String attr : settableAttrs) {
			BlockSourceCodeVisitor then = mv.visitIf("_"+attr);
			then.visitIns("(("+it.getFcItfSignature()+")attributecontroller).set"+attr+"("+attr+")");
			then.visitEnd();
		}
		mv.visitEnd();
	}

	@Override
	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, Method proxym ) {

		String mname = proxym.getName();
		String attrname = mname.substring(3);

		// Method (setter or getter) body
		if( mname.startsWith("set") ) {
			mv.visitSet(attrname,"arg0");
			mv.visitSet("_"+attrname,"true");
		}
		else {
			mv.visitIns("return",attrname);
		}
	}

	@Override
	public String getDelegatingInstance( Method proxym ) {
		/*
		 * Not relevant for this generator. Should never be called as
		 * generateProxyMethodBody is overridden by this class.
		 */
		throw new UnsupportedOperationException();
	}
}
