/***
 * Juliac
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Romain Rouvoy
 */

package org.objectweb.fractal.juliac.core.proxy;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.control.LifeCycleController;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.Interceptor;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ThenSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.lang.GenericClass;

/**
 * This class generates the source code of lifecycle interceptors.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @author Romain Rouvoy <Romain.Rouvoy@univ-lille.fr>
 */
public class LifeCycleSourceCodeGenerator
extends AbstractInterceptorSourceCodeGenerator {

	public LifeCycleSourceCodeGenerator() {
		super();
	}

	public LifeCycleSourceCodeGenerator(
		InterfaceType it, Class<?> proxycl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,proxycl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return the name of the suffix which is appended to generated class names.
	 */
	@Override
	public String getClassNameSuffix() {

		String suffix = "LC";

		if( ! mergeable ) {
			/*
			 * Append a hash code which corresponds to the name of the lifecycle
			 * controller class. The reason is that the code of each lifecycle
			 * interceptor relies on the name of the lifecycle controller class.
			 * Hence, a different interceptor class must be generated each time
			 * a new lifecycle controller class is encountered in a membrane.
			 * Although the issue does not occur with Julia, it occurs with
			 * Dream.
			 */
			String lcCtrlImplName = getLCCtrlImplName();
			int hash = lcCtrlImplName.hashCode();
			String hexhash = Integer.toHexString(hash);
			suffix += hexhash;
		}

		return suffix;
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source code generator.
	 */
	@Override
	public boolean match() {
		if( it.isFcClientItf() ) {
			return false;   // Skip client interfaces
		}
		String name = it.getFcItfName();
		if( name.endsWith("-controller") || name.equals("component") ) {
			return false;   // Skip control interfaces
		}
		return true;
	}

	@Override
	public void generateFields( ClassSourceCodeVisitor cv ) {
		if( mergeable ) {
			cv.visitField(Modifier.PUBLIC, "int", "_this_fcState", null);
			cv.visitField(Modifier.PUBLIC, "int", "_this_fcInvocationCounter", null);
		}
		else {
			String lcCtrlImplName = getLCCtrlImplName();
			cv.visitField(Modifier.PRIVATE, lcCtrlImplName, "_lc", null);
		}
	}

	@Override
	public void generateMethodInitFcController( BlockSourceCodeVisitor mv ) {
		if( ! mergeable ) {
			String lcCtrlImplName = getLCCtrlImplName();

			mv.visitVar("Object","olc","ic.getInterface(\"lifecycle-controller\")");
			BlockSourceCodeVisitor then = mv.visitIf("!(","olc","instanceof",lcCtrlImplName,")");
			BlockSourceCodeVisitor wdo = then.visitWhile("olc","instanceof",Interceptor.class.getName());
			wdo.visitSet("olc", "(("+Interceptor.class.getName()+")olc).getFcItfDelegate()").visitEnd();
			then.visitEnd();

			// Record the reference to the instance implementing the controller
			mv.visitSet("_lc","("+lcCtrlImplName+")","olc");
		}
	}

	@Override
	public void generateMethodClone( BlockSourceCodeVisitor mv ) {
		mv.visitSet("clone._lc","_lc");
	}

	@Override
	public void generateMethodOthers( ClassSourceCodeVisitor cv ) {

		if( mergeable ) {

			// public abstract void _this_incrementFcInvocationCounter()
			BlockSourceCodeVisitor mv =
				cv.visitMethod(
					Modifier.PUBLIC + Modifier.ABSTRACT, null, "void",
					"_this_incrementFcInvocationCounter", null, null );
			mv.visitEnd();

			// public abstract void _this_decrementFcInvocationCounter()
			mv =
				cv.visitMethod(
					Modifier.PUBLIC + Modifier.ABSTRACT, null, "void",
					"_this_decrementFcInvocationCounter", null, null );
			mv.visitEnd();
		}
	}

	@Override
	public void generateProxyMethodBodyBeforeCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		if( mergeable ) {
			BlockSourceCodeVisitor sync = mv.visitSync("this");
			ThenSourceCodeVisitor then = sync.visitIf("_this_fcState != 2");
			then.visitIns("_this_incrementFcInvocationCounter()");
			BlockSourceCodeVisitor _else = then.visitElse();
			_else.visitIns("_this_fcInvocationCounter++").visitEnd();
			sync.visitEnd();
		}
		else {
			BlockSourceCodeVisitor sync = mv.visitSync("_lc");
			ThenSourceCodeVisitor then = sync.visitIf("_lc.fcState != 2");
			then.visitIns("_lc.incrementFcInvocationCounter()");
			BlockSourceCodeVisitor _else = then.visitElse();
			_else.visitIns("_lc.fcInvocationCounter++").visitEnd();
			sync.visitEnd();
		}
		mv.visitln("      try {");
	}

	@Override
	public void generateProxyMethodBodyAfterReturningCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		mv.visitln("      }");
		mv.visitln("      finally {");

		if( mergeable ) {
			BlockSourceCodeVisitor sync = mv.visitSync("this");
			ThenSourceCodeVisitor then = sync.visitIf("_this_fcState != 2");
			then.visitIns("_this_decrementFcInvocationCounter()");
			BlockSourceCodeVisitor _else = then.visitElse();
			_else.visitIns("_this_fcInvocationCounter--").visitEnd();
			sync.visitEnd();
		}
		else {
			BlockSourceCodeVisitor sync = mv.visitSync("_lc");
			ThenSourceCodeVisitor then = sync.visitIf("_lc.fcState != 2");
			then.visitIns("_lc.decrementFcInvocationCounter()");
			BlockSourceCodeVisitor _else = then.visitElse();
			_else.visitIns("_lc.fcInvocationCounter--").visitEnd();
			sync.visitEnd();
		}

		mv.visitln("      }");
	}


	// ----------------------------------------------------------------------
	// Implementation specific
	// ----------------------------------------------------------------------

	/**
	 * Return the name of the class implementing the lifecycle controller.
	 */
	private String getLCCtrlImplName() {
		String signature = LifeCycleController.class.getName();
		Class<?> cl = membraneDesc.getCtrlImpl(signature);
		GenericClass uc = new GenericClass(cl);
		String name = uc.toString(cl);
		return name;
	}
}
