/***
 * Juliac
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.mergeall;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Logger;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.BindingController;
import org.objectweb.fractal.api.control.IllegalBindingException;
import org.objectweb.fractal.api.control.IllegalLifeCycleException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.BasicControllerMixin;
import org.objectweb.fractal.julia.control.binding.BasicBindingControllerMixin;
import org.objectweb.fractal.julia.control.binding.ContentBindingController;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.api.JuliacItf;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.generator.InterfaceTypeConfigurableItf;
import org.objectweb.fractal.juliac.api.generator.SourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.visit.FileSourceCodeVisitor;
import org.objectweb.fractal.juliac.core.desc.ControllerDesc;
import org.objectweb.fractal.juliac.core.desc.SimpleMembraneDesc;
import org.objectweb.fractal.juliac.core.opt.FCSourceCodeGenerator;
import org.objectweb.fractal.juliac.core.opt.FCSourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.core.opt.InitializerClassGenerator;
import org.objectweb.fractal.juliac.core.visit.FileSourceCodeWriter;
import org.objectweb.fractal.juliac.spoon.MixinClassGenerator;
import org.objectweb.fractal.juliac.spoon.helper.CtClassHelper;

import spoon.Launcher;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.factory.ClassFactory;
import spoon.reflect.factory.Factory;
import spoon.reflect.reference.CtTypeReference;
import spoon.support.compiler.VirtualFile;

/**
 * This class generates the source code associated to Fractal components.
 * Two classes are generated: the membrane implementation and the
 * initializer implementation.
 *
 * The membrane implementation is generated by merging the content, the
 * interceptors and the controllers.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 */
public class FCMergeAllSourceCodeGenerator extends FCSourceCodeGenerator {

	/** @since 2.7 */
	public static final String DEFAULT_CONFIGURATION =
		"juliac-mergeall-bundled.cfg";

	@Override
	protected void postInitJuliaLoaderModule() throws IOException {
		jloader.loadConfigFile(DEFAULT_CONFIGURATION);
		super.postInitJuliaLoaderModule();
	}


	// -----------------------------------------------------------------------
	// Implementation of the FCSourceCodeGeneratorItf interface
	// -----------------------------------------------------------------------

	public SimpleMembraneDesc generateMembraneImpl(
		ComponentType ct, String ctrldesc, String contentClassName,
		Object source )
	throws IOException {

		/*
		 * Generate business and control interface implementations.
		 */
		InterfaceType[] its = ct.getFcInterfaceTypes();
		generateInterfaceImpl(its,ctrldesc);
		InterfaceType[] membraneits = mloader.getMembraneType(ct,ctrldesc);
		generateInterfaceImpl(membraneits,ctrldesc);

		/*
		 * Retrieve interceptor generators.
		 */
		InterfaceTypeConfigurableItf icg =
			getInterceptorClassGenerator(ctrldesc);
		InterfaceTypeConfigurableItf[] iscgs =
			getInterceptorSourceCodeGenerators(ctrldesc);

		/*
		 * Prepare a membrane descriptor.
		 */
		SimpleMembraneDesc md =
			new SimpleMembraneDesc(jc,null,ctrldesc,membraneits,null,icg,iscgs);

		/*
		 * Retrieve the list of interceptor and control layers.
		 * Interceptor layers are the last ones. Last layers are executed first.
		 */
		List<CtClass<?>> layers = new ArrayList<>();
		List<CtClass<?>> interceptorLayers = getInterceptorLayers(ct,md);
		List<CtClass<?>> controlLayers = getControlLayers(ct,ctrldesc);
		layers.addAll(controlLayers);
		layers.addAll(interceptorLayers);

		/*
		 * Layers specific to components with a content.
		 */
		if( contentClassName != null ) {

			// Perform checks on the content class
			checkContentClass( contentClassName );

			List<CtClass<?>> contentCallLayers = getContentCallLayers(ct);
			layers.addAll(0,contentCallLayers);

			/*
			 * Membranes such as primitives, assume that the content provides an
			 * implementation of the BindingController interface. If this is the
			 * case, add the layer which will call this implementation.
			 */
			CtClass<?> contentBCLayer =
				getContentBindingControllerLayer(ct,layers);
			if( contentBCLayer != null ) {
				layers.add(contentBCLayer);
			}
		}

		/*
		 * Generate and register the class by merging the content, controllers
		 * and interceptors. Avoid duplicate generation.
		 */
		InitializerClassGenerator initcg =
			getInitializerClassGenerator(jc,this,null,null,null,null);
		String membraneClassName =
			initcg.getMembraneClassName(ctrldesc,contentClassName);

		if( ! jc.hasBeenGenerated(membraneClassName) ) {

			SpoonSupportItf spoon = jc.lookupFirst(SpoonSupportItf.class);
			Factory factory = spoon.getSpoonFactory();
			MixinClassGenerator spmcg = new MixinClassGenerator(factory);

			CtClass<?> generated = spmcg.generate(membraneClassName,layers);

			/*
			 * Set the superclass to be the content class.
			 */
			if( contentClassName != null ) {
				ClassFactory cf = factory.Class();
				CtTypeReference<?> ref = cf.createReference(contentClassName);
				generated.setSuperclass(ref);
			}

			/*
			 * Dump the generated class.
			 */
			spoon.generateSourceCode(generated);
		}

		/*
		 * Prepare a membrane descriptor.
		 */
		List<ControllerDesc> ctrlDescs = new ArrayList<>();
		ctrlDescs.add(new ControllerDesc(membraneClassName,null,null));
		SimpleMembraneDesc newmd =
			new SimpleMembraneDesc(
				jc,membraneClassName,ctrldesc,membraneits,ctrlDescs,icg,iscgs);

		return newmd;
	}

	public InitializerClassGenerator
	getInitializerClassGenerator(
		JuliacItf jc, FCSourceCodeGeneratorItf fcscg,
		MembraneDesc<?> membraneDesc, ComponentType ct,
		Object contentDesc, Object source ) {

		return new InitializerMergeAllClassGenerator(
			jc,fcscg,membraneDesc,ct,contentDesc,source);
	}


	// -----------------------------------------------------------------------
	// Implementation specific
	// -----------------------------------------------------------------------

	/**
	 * Perform checks on the content class.
	 */
	private void checkContentClass( String contentClassName )
	throws IOException {

		/*
		 * Check that the binding control methods in the content class declare
		 * in their throws clause NoSuchInterfaceException,
		 * IllegalBindingException and IllegalLifeCycleException. In the merge
		 * all mode, the membrane class extends the content class, and the Java
		 * compiler complains if the content class does not throw these
		 * exceptions.
		 *
		 * If the content class does not throw these exceptions, report a
		 * warning and generate runtime exceptions.
		 */

		Class<?> bcuc = jc.loadClass(BindingController.class.getName());
		Class<?> cl = jc.loadClass(contentClassName);
		if( ! bcuc.isAssignableFrom(cl) ) {
			return;
		}

		Method mbind=null, mlookup=null, munbind=null;
		try {
			mbind =
				cl.getMethod(
					"bindFc", new Class<?>[]{String.class,Object.class} );
			mlookup = cl.getMethod("lookupFc", new Class<?>[]{String.class});
			munbind = cl.getMethod("unbindFc", new Class<?>[]{String.class});
		}
		catch (NoSuchMethodException e) {
			// As cl implements BindingController, these methods should be found
			throw new JuliacRuntimeException(e);
		}

		Method[] meths = new Method[]{mbind,mlookup,munbind};
		checkMethodsThrowsException(
			contentClassName, meths, NoSuchInterfaceException.class );

		meths = new Method[]{mbind,munbind};
		checkMethodsThrowsException(
			contentClassName, meths, IllegalBindingException.class );
		checkMethodsThrowsException(
			contentClassName, meths, IllegalLifeCycleException.class );
	}

	/**
	 * Check that the specified methods declare the specified exception.
	 *
	 * If this is not the case, generate the source code of a {@link
	 * RuntimeException} for the specified exception.
	 */
	private void checkMethodsThrowsException(
		String contentClassName, Method[] meths,
		Class<? extends Throwable> e )
	throws IOException {

		String ms = "";
		for (Method meth : meths) {
			boolean t = checkThrowsException(meth,e);
			if(!t) ms+= meth.getName()+" ";
		}

		if( ms.length() > 0 ) {
			final String msg =
				"Method(s) "+ms+"in class "+contentClassName+ " should throw "+
				e.getName()+
				". This exception will be weaken to a RuntimeException.";
			Logger logger = jc.getLogger();
			logger.warning(msg);

			String targetClassName = e.getName();
			SourceCodeGeneratorItf scg =
				new RuntimeExceptionClassGenerator(targetClassName);
			jc.generateSourceCodeOverride(scg);
		}
	}

	/**
	 * Return <code>true</code> if the specified method declares the specified
	 * exception in its throws clause.
	 */
	private boolean checkThrowsException(
		Method m, Class<? extends Throwable> e ) {

		Class<?>[] exceptions = m.getExceptionTypes();
		for (Class<?> exception : exceptions) {
			if( e.isAssignableFrom(exception) ) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Return the list of mixin layers which call the content class (super
	 * call).
	 */
	private List<CtClass<?>> getContentCallLayers( ComponentType ct ) {

		SpoonSupportItf spoon = jc.lookupFirst(SpoonSupportItf.class);
		Factory factory = spoon.getSpoonFactory();

		ContentCallCtClassGenerator ccscg =
			new ContentCallCtClassGenerator(factory);

		List<CtClass<?>> contentCallLayers = new ArrayList<>();
		InterfaceType[] its = ct.getFcInterfaceTypes();
		for (InterfaceType it : its) {

			if( it.isFcClientItf() ) {
				continue;   // Skip client interface types
			}

			String signature = it.getFcItfSignature();
			Class<?> cl = jc.loadClass(signature);
			CtClass<?> contentCallLayer = ccscg.generateCtClass(it,cl);
			contentCallLayers.add(contentCallLayer);
		}

		return contentCallLayers;
	}

	/**
	 * Return the list of mixin layers which implement the interceptor code for
	 * the specified component type and the specified membrane descriptor.
	 */
	private List<CtClass<?>> getInterceptorLayers(
		ComponentType ct, MembraneDesc<?> membraneDesc )
	throws IOException {

		List<CtClass<?>> interceptorLayers = new ArrayList<>();

		InterfaceTypeConfigurableItf[] iscgs =
			membraneDesc.getInterceptorSourceCodeGenerators();
		if( iscgs == null ) {
			// No interceptor
			return interceptorLayers;
		}

		for (InterfaceTypeConfigurableItf iscg : iscgs) {

			iscg.setMembraneDesc(membraneDesc);
			iscg.setMergeable(true);

			/*
			 * The set of interface signatures for which an interception layer
			 * has already been generated.
			 */
			Set<String> signatures = new HashSet<>();

			InterfaceType[] its = ct.getFcInterfaceTypes();
			for (InterfaceType it : its) {

				// Skip client and attribute-controller
				if( it.isFcClientItf() ||
					it.getFcItfName().equals("attribute-controller") ) {
					continue;
				}

				// Skip if a layer has already been generated for this signature
				String signature = it.getFcItfSignature();
				if( signatures.contains(signature) ) {
					continue;
				}

				/*
				 * Create the interceptor class generator.
				 */
				Class<?> cl = jc.loadClass(signature);
				final String pkgRoot = jc.getPkgRoot();
				iscg.setInterfaceType(it,cl,pkgRoot);

				/*
				 * Generate the code of the interceptor.
				 */
				StringWriter sw = new StringWriter();
				PrintWriter pw = new PrintWriter(sw);
				FileSourceCodeVisitor fw = new FileSourceCodeWriter(pw);
				iscg.generate(fw);
				pw.close();
				sw.close();
				String interceptorSnippet = sw.toString();

				/*
				 * Remove the header and the package declaration in order for
				 * the noclasspath mode of Spoon to generate a correct model of
				 * the interceptor class. Else the signature of the implemented
				 * interface is wrongly associated with the same package as the
				 * one of the interceptor class.
				 */
				int idx = interceptorSnippet.indexOf("abstract");
				interceptorSnippet = interceptorSnippet.substring(idx);
				Launcher launcher = new Launcher();
				launcher.addInputResource(new VirtualFile(interceptorSnippet));
				launcher.getEnvironment().setNoClasspath(true);
				CtClass<?> interceptorLayer =
					(CtClass<?>) launcher.buildModel().getAllTypes().iterator().next();

				interceptorLayers.add(interceptorLayer);
				signatures.add(signature);
			}
		}

		return interceptorLayers;
	}

	/**
	 * Return the layers associated with the specified control descriptor.
	 */
	private List<CtClass<?>> getControlLayers(
		ComponentType ct, String ctrlDesc ) {

		List<String> layers = new UniqueArrayList<>();

		List<ControllerDesc> cdescs = mloader.getCtrlImplLayers(ct,ctrlDesc);
		for (ControllerDesc cdesc : cdescs) {
			List<String> cdescLayers = cdesc.getLayerImplNames();
			if( cdescLayers == null ) {
				/*
				 * Special case where the name of the controller implementation
				 * is directly specified (not generated with
				 * MixinClassGenerator).
				 */
				String cdescImplName = cdesc.getImplName();
				layers.add(cdescImplName);
			}
			else {
				layers.addAll(cdescLayers);
			}
		}

		String bcmName = BasicControllerMixin.class.getName();
		if( layers.contains(bcmName) ) {
			/*
			 * Julia assumes that BasicControllerMixin is always the first layer
			 * for controllers. This layer does not propagate calls to
			 * initFcController. Julia reports an error if one tries to modify
			 * this (not super method). The policy implemented by
			 * UniqueArrayList for managing duplicate layers may lead to
			 * cases where BasicControllerMixin is not the first layer if it
			 * appears several times (only the last occurrence is kept.) In
			 * order to deal with such a situation, reintroduce
			 * BasicControllerMixin as the first layer.
			 */
			layers.add(0,bcmName);
		}

		SpoonSupportItf spoon = jc.lookupFirst(SpoonSupportItf.class);
		Factory factory = spoon.getSpoonFactory();
		List<CtClass<?>> ctLayers = CtClassHelper.toCtClass(factory,layers);

		return ctLayers;
	}

	/**
	 * Control membranes such as primitive ones expect that the content class
	 * provides an implementation for the {@link BindingController} interface.
	 * When this is the case, a layer implementing the
	 * {@link ContentBindingController} must be provided. This method returns
	 * this layer or <code>null</code> if the layer is not needed.
	 */
	private CtClass<?> getContentBindingControllerLayer(
		ComponentType ct, List<CtClass<?>> layers ) {

		SpoonSupportItf spoon = jc.lookupFirst(SpoonSupportItf.class);
		Factory factory = spoon.getSpoonFactory();

		CtClass<BasicBindingControllerMixin> ctbbcm =
			factory.Class().get(BasicBindingControllerMixin.class);

		if( ! isAssignableFrom(layers,BindingController.class) ||
			layers.contains(ctbbcm) ||
			! containsClientInterfaceType(ct) ) {

			return null;
		}

		/*
		 * ContentBindingControllerSourceCodeGenerator generates a default
		 * implementation of the ContentBindingController interface. This
		 * implementation is expected by control membranes such as
		 * primitive ones, which expect that their content class provides an
		 * implementation of the BindingController interface (if the component
		 * defines a client interface).
		 */

		ContentBindingControllerCtClassGenerator cbcscg =
			new ContentBindingControllerCtClassGenerator(factory);

		InterfaceType it =
			new BasicInterfaceType(
				"cbci",ContentBindingController.class.getName(),
				false,false,false);
		CtClass<?> generated =
			cbcscg.generateCtClass(it,ContentBindingController.class);
		return generated;
	}

	/**
	 * Return <code>true</code> if the specified component type defines a client
	 * interface.
	 */
	private static boolean containsClientInterfaceType( ComponentType ct ) {
		InterfaceType[] its = ct.getFcInterfaceTypes();
		for (InterfaceType it : its) {
			if( it.isFcClientItf() ) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Return <code>true</code> if the specified interface is assignable from
	 * one of the classes of the specified list.
	 */
	private boolean isAssignableFrom( List<CtClass<?>> layers, Class<?> itf ) {

		SpoonSupportItf spoon = jc.lookupFirst(SpoonSupportItf.class);
		Factory factory = spoon.getSpoonFactory();
		CtTypeReference<?> itftr = factory.Type().createReference(itf);

		for (CtClass<?> layer : layers) {
			CtTypeReference<?> layertr = layer.getReference();
			if( layertr.isSubtypeOf(itftr) ) {
				return true;
			}
		}
		return false;
	}
}
