/***
 * Juliac
 * Copyright (C) 2009-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.osgi;

import java.util.Iterator;
import java.util.Set;
import java.util.TreeSet;

import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.runtime.JuliacBootstrapComponentImpl;
import org.objectweb.fractal.juliac.runtime.RuntimeException;

/**
 * The Juliac bootstrap component for OSGi.
 *
 * This bootstrap component provides a type-factory and a generic-factory for
 * instantiating osgiPrimitive components.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.2
 */
public class JuliacOSGiBootstrapComponentImpl
extends JuliacBootstrapComponentImpl {

	/**
	 * <p>
	 * Analyze the controller descriptor specified as the second argument of
	 * {@link #newFcInstance(Type, Object, Object)}.
	 * </p>
	 *
	 * @param controllerDesc  the controller descroptor to analyze
	 * @return                the controller descriptor as a string
	 */
	protected String checkControllerDesc( Object controllerDesc )
	throws InstantiationException {

		/*
		 * We really need to call super#checkControllerDesc first in order to
		 * use the correct controller descriptor in the case controllerDesc
		 * contains some hints in an array.
		 */
		String ctrlDesc = super.checkControllerDesc(controllerDesc);

		if( ! ctrlDesc.equals("osgiPrimitive") ) {
			String msg = "Only osgiPrimitive components are supported";
			throw new InstantiationException(msg);
		}

		return ctrlDesc;
	}

	/**
	 * Return the name of the factory class associated with the specified {@link
	 * ComponentType}, controller descriptor and content descriptor.
	 *
	 * This name has been generated by {@link
	 * org.objectweb.fractal.juliac.felix.InitializerOOCtrlClassGenerator#getTargetClassName()}.
	 */
	protected String getFactoryClassName(
		ComponentType ct, String ctrlDesc, Object contDesc ) {

		/*
		 * In the case of osgiPrimitive component, content descriptors are of
		 * the form 'bundle1 bundle2 ... bundlen', e.g.
		 * osgi-felix-helloworld-interfaces.jar osgi-felix-helloworld-server.jar
		 */
		StringBuilder sb = new StringBuilder();
		String contentDescStr = (String) contDesc;
		int len = contentDescStr.length();
		for (int i = 0; i < len; i++) {
			char c = contentDescStr.charAt(i);
			// TODO we should also replace all characters which are not legal in package names.
			if( c==' ' || c=='-' ) {c='.';}
			sb.append(c);
		}
		String contentClassName = sb.toString();

		/*
		 * Compute the hash code used as the suffix by the initializer class.
		 * The idea is that there is a class for each different record of
		 * content class, controller descriptor and component type. The two
		 * first elements can act, by nature, as identity. A hash is computed to
		 * uniquely identify component types. Morevover interface types are
		 * sorted. See the comment in {@link
		 * org.objectweb.fractal.juliac.Utils#hexhash(ComponentType)} for the
		 * rationale for this choice.
		 */
		InterfaceType[] its = ct.getFcInterfaceTypes();
		Set<String> names = new TreeSet<>();
		for (int i = 0; i < its.length; i++) {
			InterfaceType it = its[i];
			String name = it.getFcItfName();
			names.add(name);
		}

		StringBuilder sct = new StringBuilder("[");
		boolean first = true;
		for (Iterator<String> iterator = names.iterator(); iterator.hasNext();) {
			String name = iterator.next();
			InterfaceType it = null;
			try {
				it = ct.getFcInterfaceType(name);
			}
			catch (NoSuchInterfaceException nsie) {
				// Shouldn't occur since name if one the interface names from ct
				throw new RuntimeException(nsie);
			}
			if(first) { first = false; }
			else { sct.append(','); }
			sct.append(it.toString());
		}
		sct.append(']');

		/*
		 * toString() mandatory to compute the hash on the String, not the
		 * StringBuilder. 2 Strings with the same content have the same hash
		 * code, whereas 2 StringBuilders with the same content are considered as
		 * different objects, and do not share the same hash code.
		 */
		int hashct = sct.toString().hashCode();
		String hexhashct = Integer.toHexString(hashct);

		String membraneClassName = contentClassName+"FC"+ctrlDesc;
		String factoryClassName = membraneClassName+"FC"+hexhashct;

		return factoryClassName;
	}
}
