/***
 * Juliac
 * Copyright (C) 2010-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.osgi;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.jar.Attributes;
import java.util.jar.JarEntry;
import java.util.jar.JarOutputStream;
import java.util.jar.Manifest;

import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.control.AttributeController;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.JuliacItf;
import org.objectweb.fractal.juliac.api.generator.ProxyClassGeneratorItf;
import org.objectweb.fractal.juliac.api.generator.SourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.core.proxy.InterfaceImplementationClassGenerator;

/**
 * Utility source code generator for membranes which control OSGi components.
 *
 * This generator is used by {@link FCOOCtrlSourceCodeGenerator} for
 * <code>osgiPrimitive</code> and <code>osgComposite</code> membranes and in
 * Tinfi, by {@link org.ow2.frascati.tinfi.osgi.FCOOCtrlSourceCodeGenerator} for
 * <code>scaOsgiPrimitive</code> membranes.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.2.5
 */
public class FCOOCtrlGenerator {

	private JuliacItf jc;

	public FCOOCtrlGenerator( JuliacItf jc ) {
		this.jc = jc;
	}

	/**
	 * Check that the specified parameters are suitable for membranes which
	 * encapsulate OSGi components.
	 *
	 * @param type            the component type
	 * @param controllerDesc  the component controller descriptor
	 * @param contentDesc     the component content descriptor
	 * @throws IllegalArgumentException  if the parameters are not suitable
	 */
	public void check( Type type, Object controllerDesc, Object contentDesc ) {

		/*
		 * Check parameter consistence. It's safe to perform the casts after
		 * having performed the checks.
		 */
		if( !(contentDesc instanceof String) ) {
			String msg = "contentDesc should be a String";
			throw new IllegalArgumentException(msg);
		}
		String contentDescStr = (String) contentDesc;

		/*
		 * Content descriptors are of the form 'bundle1 bundle2 ... bundlen',
		 * e.g. osgi-felix-helloworld-interfaces.jar osgi-felix-helloworld-server.jar
		 */
		String[] jars = contentDescStr.split(" ");
		for (String jar : jars) {
			if( ! jar.endsWith(".jar") ) {
				String msg =
					jar+" in '"+contentDescStr+
					"' is not a .jar ending file name";
				throw new IllegalArgumentException(msg);
			}
			if( jar.length() <= 4 ) {
				String msg = jar+" in '"+contentDescStr+
				"' should define a file name";
				throw new IllegalArgumentException(msg);
			}
		}
	}

	/**
	 * Return the source code generator for component interfaces.
	 */
	public ProxyClassGeneratorItf getInterfaceClassGenerator(InterfaceType it) {

		final String signature = it.getFcItfSignature();
		final Class<?> cl = jc.loadClass(signature);
		final String pkgRoot = jc.getPkgRoot();

		/*
		 * If the interface is an attribute control interface, generate an
		 * implementation which uses dynamic method invocations for delegating
		 * to the content instance. We don't need this for server business
		 * interfaces, since in this case, the delegation is performed by
		 * lifecycle interceptors generated by DynInterceptorClassGenerator.
		 *
		 * No need to take care of cloneable attribute control interfaces since
		 * these interfaces are used for parametric components. We don't have
		 * any such components in this optimization level (only osgiPrimitive.)
		 */

		ProxyClassGeneratorItf pcg =
			AttributeController.class.isAssignableFrom(cl) ?
			new DynInterfaceImplementationClassGenerator(it,cl,pkgRoot,false) :
			new InterfaceImplementationClassGenerator(it,cl,pkgRoot,false);

		return pcg;
	}

	/**
	 * Generate a bundle for the proxy from the OSGi world to the Fractal world
	 * whose name and interface type are specified
	 */
	private void generateBundle( String serviceClassName, InterfaceType it )
	throws IOException {

		// Compile the code to get a binary version of serviceClassName
		jc.build();

		// Jar file name
		String signature = it.getFcItfSignature();
		String jarfname = signature.replace('.','-');

		int lastdot = serviceClassName.lastIndexOf('.');
		String pkgName = serviceClassName.substring(0,lastdot);
		lastdot = signature.lastIndexOf('.');
		String sigPkgName = signature.substring(0,lastdot);

		// Create the manifest
		Manifest man = new Manifest();
		Attributes attrs = man.getMainAttributes();
		attrs.put(Attributes.Name.MANIFEST_VERSION,"1.0");
		attrs.putValue("Bundle-Activator",pkgName+".Activator");
		attrs.putValue("Bundle-ManifestVersion","2");
		attrs.putValue("Bundle-Name",jarfname);
		attrs.putValue("Bundle-SymbolicName",jarfname);
		attrs.putValue("Export-Package",pkgName);
		StringBuilder sb = new StringBuilder();
		sb.append(sigPkgName);
		sb.append(',');
		sb.append("org.objectweb.fractal.api");
		sb.append(',');
		sb.append("org.objectweb.fractal.api.factory");
		sb.append(',');
		sb.append("org.objectweb.fractal.julia");
		sb.append(',');
		sb.append("org.objectweb.fractal.juliac.runtime");
		sb.append(',');
		sb.append("org.osgi.framework");
		sb.append(',');
		sb.append(pkgName);
		attrs.putValue( "Import-Package", sb.toString() );

		// Create the jar file
		File classDir = jc.getClassDir();
		File jarFile = new File(classDir,jarfname+".jar");
		FileOutputStream fos = new FileOutputStream(jarFile);
		JarOutputStream jos = new JarOutputStream(fos,man);

		// Get the .class file corresponding the serviceClassName
		String classfname = serviceClassName.replace('.','/');
		classfname += ".class";
		File classFile = new File(classDir,classfname);
		putNextEntry(jos,classFile,classfname);

		// Get the corresponding Activator class
		classfname = pkgName+".Activator";
		classfname = classfname.replace('.','/');
		classfname += ".class";
		classFile = new File(classDir,classfname);
		putNextEntry(jos,classFile,classfname);

		jos.close();
		fos.close();
	}

	/**
	 * Generate the source code of component interfaces associated with the
	 * specified interface type.
	 */
	public void generateInterfaceImpl( InterfaceType it )
	throws IOException {

		/*
		 * No need to generate attribute controller implementations like in the
		 * overriden method. These implementations are used for template
		 * components. We don't have any such components in this optimization
		 * level (only osgiPrimitive.)
		 */
		SourceCodeGeneratorItf pcg = getInterfaceClassGenerator(it);
		jc.generateSourceCode(pcg);

		/*
		 * For client interfaces, generate a proxy for transmitting invocations
		 * from the OSGi world to the Fractal world.
		 */
		if( it.isFcClientItf() ) {

			// Proxy generation
			final String signature = it.getFcItfSignature();
			final Class<?> cl = jc.loadClass(signature);
			final String pkgRoot = jc.getPkgRoot();
			SourceCodeGeneratorItf dscg =
				new DynServiceClassGenerator(it,cl,pkgRoot,null,false);
			jc.generateSourceCode(dscg);

			// Generate an Activator class for the proxy
			String targetClassName = dscg.getTargetTypeName();
			int lastdot = targetClassName.lastIndexOf('.');
			String pkgName = targetClassName.substring(0,lastdot);
			SourceCodeGeneratorItf aicg =
				new DynServiceActivatorClassGenerator(
					pkgName, targetClassName, it );
			jc.generateSourceCode(aicg);

			// Create the bundle for the proxy and the Activator
			generateBundle(targetClassName,it);
		}
	}


	// ------------------------------------------------------------------
	// Implementation specific
	// ------------------------------------------------------------------

	/**
	 * Put a next entry named <code>'entryName</code>in the specified {@link
	 * JarOutputStream} and append the binary content of the specified file.
	 */
	private static void putNextEntry(
		JarOutputStream jos, File file, String entryName )
	throws IOException {

		try( FileInputStream fis = new FileInputStream(file) ) {

			JarEntry je = new JarEntry(entryName);
			jos.putNextEntry(je);

			byte[] buffer = new byte[1024];
			int length = 0;
			while( (length=fis.read(buffer)) != -1 ) {
				jos.write(buffer,0,length);
			}

			jos.closeEntry();
		}
	}
}
