/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2011-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.opt.oo.light;

import java.lang.annotation.Annotation;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Vector;

import org.oasisopen.sca.ComponentContext;
import org.oasisopen.sca.ServiceReference;
import org.oasisopen.sca.annotation.Property;
import org.oasisopen.sca.annotation.Scope;
import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.NoSuchInterfaceException;
import org.objectweb.fractal.api.control.NameController;
import org.objectweb.fractal.api.factory.InstantiationException;
import org.objectweb.fractal.api.type.ComponentType;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.fraclet.types.Contingency;
import org.objectweb.fractal.julia.InitializationContext;
import org.objectweb.fractal.juliac.api.JuliacItf;
import org.objectweb.fractal.juliac.api.generator.SourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.CatchSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ThenSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.ipf.CompositeInjectionPointHashMap;
import org.objectweb.fractal.juliac.commons.ipf.InjectionPoint;
import org.objectweb.fractal.juliac.commons.ipf.InjectionPointFieldImpl;
import org.objectweb.fractal.juliac.commons.ipf.InjectionPointMethodImpl;
import org.objectweb.fractal.juliac.commons.lang.ClassHelper;
import org.objectweb.fractal.juliac.commons.lang.annotation.AnnotationHelper;
import org.objectweb.fractal.juliac.commons.lang.reflect.MemberHelper;
import org.objectweb.fractal.juliac.core.Constants;
import org.objectweb.fractal.juliac.core.opt.ClassGenerator;
import org.ow2.frascati.tinfi.CallbackManager;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.api.control.ContentInstantiationException;
import org.ow2.frascati.tinfi.control.content.ContentClassMetaData;
import org.ow2.frascati.tinfi.control.content.IllegalContentClassMetaData;
import org.ow2.frascati.tinfi.oasis.ServiceReferenceImpl;

/**
 * The class generates implementations of the SCA Content Controller specific
 * for a given content class.
 *
 * @author Lionel <Lionel.Seinturier@univ-lille.fr>
 * @since 1.4.3
 */
public abstract class SCAContentControllerClassGenerator
extends ClassGenerator {

	/** Suffix for classes generated by this generator. */
	public static final String SUFFIX = "FCSCAContentControllerImpl";

	private JuliacItf jc;
	protected String contentClassName;
	protected ContentClassMetaData ccmd;

	/**
	 * Factory method for {@link SCAContentControllerClassGenerator}s. Depending
	 * on the @{@link Scope} annotation of the specified content class, an
	 * instance of one of the subclasses of {@link
	 * SCAContentControllerClassGenerator} is returned.
	 *
	 * @param contentClassName  the content class
	 * @throws IllegalContentClassMetaData
	 *         if the specified content class does not respect the rules defined by
	 *         the SCA Java implementation type standard, e.g. @Init annotated
	 *         methods must not have parameters, etc.
	 * @throws IllegalArgumentException  if the scope is not supported
	 * @since 1.4.5
	 */
	public static SourceCodeGeneratorItf get( JuliacItf jc, String contentClassName )
	throws IllegalContentClassMetaData {

		Class<?> cl = jc.loadClass(contentClassName);
		ContentClassMetaData ccmd = ContentClassMetaData.get(cl);

		SourceCodeGeneratorItf scacccg = null;
		String scope = ccmd.scope==null ? null : ccmd.scope.toUpperCase();
		if( scope == null || scope.equals("STATELESS") ) {
			scacccg =
				new SCAContentControllerStatelessScopeClassGenerator(
						jc,contentClassName,ccmd);
		}
		else if( scope.equals("COMPOSITE") ) {
			scacccg =
				new SCAContentControllerCompositeScopeClassGenerator(
						jc,contentClassName,ccmd);
		}
		else {
			final String msg = "Unsupported scope: "+ccmd.scope;
			throw new IllegalArgumentException(msg);
		}

		return scacccg;
	}

	protected SCAContentControllerClassGenerator(
		JuliacItf jc, String contentClassName, ContentClassMetaData ccmd ) {

		this.jc = jc;
		this.contentClassName = contentClassName;
		this.ccmd = ccmd;
	}

	/**
	 * Return the name of the class generated by this generator.
	 */
	public String getTargetTypeName() {
		final String pkgRoot = jc.getPkgRoot();
		final String targetClassName = pkgRoot + contentClassName + SUFFIX;
		return targetClassName;
	}

	@Override
	public String getSuperClassName() {
		return Constants.JULIAC_RUNTIME_GENERATED+".SCAContentControllerImpl";
	}

	@Override
	public void generateFields( ClassSourceCodeVisitor cv ) {
		generateFieldContent(cv);
		cv.visitField(Modifier.PRIVATE,"boolean","instantiatedOnce","false");
		cv.visitField(
			Modifier.PRIVATE+Modifier.STATIC,Hashtable.class.getName(),"props",
			null);
		cv.visitField(
			Modifier.PRIVATE+Modifier.STATIC,"String[]","propnames",null);
	}

	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {
		generateMethodInitFcController(cv);
		generateMethodGetFcContentClass(cv);
		generateMethodGetFcContent(cv);
		generateMethodReleaseFcContent(cv);
		generateMethodGetFcCurrentContents(cv);
		generateMethodSetPropertyValue(cv);
		generateMethodGetPropertyValue(cv);
		generateMethodSetReferenceValue(cv);
		generateMethodUnsetReferenceValue(cv);
		generateMethodEagerInit(cv);
		generateMethodStart(cv);
		generateMethodStop(cv);
		generateMethodContainsPropertyName(cv);
		generateMethodGetPropertyNames(cv);
		generateMethodGetPropertyType(cv);
	}

	protected abstract void generateFieldContent( ClassSourceCodeVisitor cv );
	protected abstract void generateMethodReleaseFcContent( ClassSourceCodeVisitor cv );


	// --------------------------------------------------------------------
	// Generation of the initFcController method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodInitFcController( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,"void","initFcController",
				new String[]{InitializationContext.class.getName()+" ic"},
				new String[]{InstantiationException.class.getName()});
		mv.visitIns("super.initFcController(ic)");

		Map<String,Class<?>> props = getProperties();
		ThenSourceCodeVisitor thenv = mv.visitIf("props == null");

		thenv.visitSet("props","new ",Hashtable.class.getName(),"()");
		thenv.visitSet("propnames","new String[",props.size(),"]");

		// Generate property registrations
		int i = 0;
		for (Map.Entry<String,Class<?>> entry : props.entrySet()) {
			String name = entry.getKey();
			Class<?> type = entry.getValue();
			String tname = ClassHelper.getName(type);
			mv.visitIns("props.put(\""+name+"\",",tname,".class)");
			mv.visitSet("propnames["+i+"]","\""+name+"\"");
			i++;
		}

		thenv.visitEnd();
		mv.visitEnd();
	}

	// --------------------------------------------------------------------
	// Generation of the getFcContentClass method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodGetFcContentClass( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,"Class","getFcContentClass",null,null);
		mv.visitIns("return",ccmd.fcContentClass.getName(),".class");
		mv.visitEnd();
	}

	// --------------------------------------------------------------------
	// Generation of the getFcContent method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodGetFcContent( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,contentClassName,"getFcContent",null,
				new String[]{ContentInstantiationException.class.getName()});
		generateMethodGetFcContent(mv);
		mv.visitEnd();
	}

	/**
	 * @since 1.4.5
	 */
	protected abstract void generateMethodGetFcContent( BlockSourceCodeVisitor mv );

	protected void generateCreateFcInstance( BlockSourceCodeVisitor mv ) {

		if( ccmd.constructorAnnotatedElement != null ) {

			/*
			 * @Constructor annotated constructor.
			 * Retrieve @Property annotated parameters.
			 */
			List<String> names = new ArrayList<>();
			List<Class<?>> types = new ArrayList<>();
			retrieveInitParameters(ccmd.constructorAnnotatedElement,names,types);

			if( names.size() == 0 ) {
				mv.visitSet("content","new",contentClassName,"()");
			}
			else {
				StringBuilder sb = new StringBuilder();
				for (String name : names) {
					sb.append(",\"");
					sb.append(name);
					sb.append('"');
				}
				mv.visitSet("Object[] values","getInitValues(\"",contentClassName,"\"",sb.toString(),")");

				sb = new StringBuilder();
				boolean first = true;
				for (int i = 0; i < names.size(); i++) {
					if(first) { first=false; } else { sb.append(','); }
					sb.append('(');
					Class<?> type = types.get(i);
					sb.append(type.getName());
					sb.append(")values[");
					sb.append(i);
					sb.append(']');
				}
				mv.visitSet("content","new",contentClassName,"(",sb.toString(),")");
			}
		}
		else if( ccmd.providesMethod != null ) {
			// @Provides annotated factory method
			mv.visitSet(contentClassName+" obj","new",contentClassName,"()");
			mv.visitSet("content","obj.",ccmd.providesMethod.getName(),"()");
		}
		else {
			// Regular content class instantiation
			mv.visitSet("content","new",contentClassName,"()");
		}
	}

	protected void generateInitializeFcInstance( BlockSourceCodeVisitor mv ) {

		// Inject control interfaces
		boolean first = true;
		for (InjectionPoint<Annotation> ip : ccmd.controllerAnnotatedElements) {

			Annotation annot = ip.getAnnotation();
			Class<?> type = ip.getType();
			String name = AnnotationHelper.getAnnotationParamValue(annot,"name");

			/*
			 * Do not try to optimize by taking the value of the weaveableC
			 * field when the component control interface is requested. The
			 * reason follows.
			 *
			 * The value stored in comp has been set when mixing the layers and
			 * originates from the getInterface() method of
			 * InitializationContext. This method skips the component interface
			 * and returns the delegate. Here the delegate is the reference of
			 * the interceptor. It is much better for the programmer to retrieve
			 * in the injection point a reference which can be cast to
			 * Interface. We then query comp to retrieve the "real" component
			 * interface.
			 */
			String decl = "";
			if(first) {
				first = false;
				mv.visitSet("Object ctrlItf","null");
				decl = "Class<?> ";
			}
			CatchSourceCodeVisitor catchv = mv.visitTry();
			catchv.visitSet("ctrlItf","weaveableC.getFcInterface(\""+name+"\")");
			catchv = catchv.visitCatch(NoSuchInterfaceException.class,"nsie");
			catchv.visitIns("throw new",ContentInstantiationException.class.getName(),"(nsie)");
			catchv.visitEnd();
			mv.visitSet(decl+"ctrlItfCl","ctrlItf.getClass()");
			ThenSourceCodeVisitor thenv =
				mv.visitIf("!",type.getName(),".class.isAssignableFrom(ctrlItfCl)");
			thenv.visitSet("String msg","\"The type \"+ctrlItfCl.getName()+\" of the "+name+" control interface can not be injected into ",type.getName()+"\"");
			thenv.visitIns("throw new ",ContentInstantiationException.class.getName(),"(msg)");
			thenv.visitEnd();

			generateSetInjectionPointValue(mv,ip,"ctrlItf");
		}

		// Inject the component context
		if( ccmd.contextAnnotatedElement != null ) {
			CatchSourceCodeVisitor catchv = mv.visitTry();
			catchv.visitSet(ComponentContext.class.getName()+" compCtx","("+ComponentContext.class.getName()+")weaveableC.getFcInterface(\"sca-component-controller\")");
			generateSetInjectionPointValue(catchv,ccmd.contextAnnotatedElement,"compCtx");
			catchv = catchv.visitCatch(NoSuchInterfaceException.class,"nsie");
			catchv.visitIns("throw new",ContentInstantiationException.class.getName(),"(nsie)");
			catchv.visitEnd();
		}

		// Inject the component name
		if( ccmd.componentNameAnnotatedElement != null ) {
			CatchSourceCodeVisitor catchv = mv.visitTry();
			catchv.visitSet(NameController.class.getName()+" nc","("+NameController.class.getName()+")weaveableC.getFcInterface(\"name-controller\")");
			catchv.visitSet("String name","nc.getFcName()");
			generateSetInjectionPointValue(catchv,ccmd.componentNameAnnotatedElement,"name");
			catchv = catchv.visitCatch(NoSuchInterfaceException.class,"nsie");
			catchv.visitIns("throw new",ContentInstantiationException.class.getName(),"(nsie)");
			catchv.visitEnd();
		}

		// Inject annotated references
		for (Map.Entry<String,InjectionPoint<Annotation>> entry : ccmd.refs.entrySet()) {
			String name = entry.getKey();
			InjectionPoint<Annotation> ip = entry.getValue();
			Class<?> iptype = ip.getType();
			boolean isRequired = isRequired(ip);

			if( List.class.isAssignableFrom(iptype) ||
				Vector.class.isAssignableFrom(iptype) ) {

				// SCA style multiple reference
				mv.visitSet("java.util.List srs","getCollectionInterfacesAsList(\""+name+"\")");
				generateSetInjectionPointValue(mv,ip,"srs");
				if(isRequired){
					BlockSourceCodeVisitor forv = mv.visitFor("int i=0 ; i<srs.size() ; i++");
					forv.visitSet("Object o","srs.get(i)");
					ThenSourceCodeVisitor thenv = forv.visitIf("o == null");
					thenv.visitSet("String msg","\"Required reference "+name+" in component class "+contentClassName+" should have been set\"");
					thenv.visitIns("throw new",ContentInstantiationException.class.getName(),"(msg)");
					thenv.visitEnd();
					forv.visitEnd();
				}
			} else if( Map.class.isAssignableFrom(iptype) ||
					   Hashtable.class.isAssignableFrom(iptype) ) {

				// Fraclet style multiple reference
				mv.visitSet("java.util.Map map","getCollectionInterfacesAsMap(\""+name+"\")");
				generateSetInjectionPointValue(mv,ip,"map");
				if(isRequired){
					mv.visitSet("java.util.Collection srs","map.values()");
					BlockSourceCodeVisitor forv = mv.visitFor("Object o : srs");
					ThenSourceCodeVisitor thenv = forv.visitIf("o == null");
					thenv.visitSet("String msg","\"Required reference "+name+" in component class "+contentClassName+" should have been set\"");
					thenv.visitIns("throw new",ContentInstantiationException.class.getName(),"(msg)");
					thenv.visitEnd();
					forv.visitEnd();
				}
			}
			else {
				// Singleton reference
				CatchSourceCodeVisitor catchv = mv.visitTry();
				catchv.visitSet("Object itf","weaveableC.getFcInterface(\""+name+"\")");
				catchv.visitSet("Object sr","getServiceReference(itf)");
				generateSetInjectionPointValue(mv,ip,"sr");
				if(isRequired) {
					ThenSourceCodeVisitor thenv = mv.visitIf("sr == null");
					thenv.visitSet("String msg","\"Required reference "+name+" in component class "+contentClassName+" should have been set\"");
					thenv.visitIns("throw new",ContentInstantiationException.class.getName(),"(msg)");
					thenv.visitEnd();
				}
				catchv = catchv.visitCatch(NoSuchInterfaceException.class,"nsie");
				/*
				 * The annotated reference is not declared in the component
				 * type. This corresponds to an annotated class that is used
				 * with different component type. No value needs to be injected.
				 */
				catchv.visitEnd();
			}
		}

		// Inject annotated properties
		first = true;
		for (Map.Entry<String,InjectionPoint<Annotation>> entry : ccmd.props.entrySet()) {
			String name = entry.getKey();
			InjectionPoint<Annotation> ip = entry.getValue();

			String decl = "";
			if(first) { decl="boolean "; first=false; }
			mv.visitSet(decl+"hasBeenSet","weaveableSCAPC.hasBeenSet(\""+name+"\")");
			ThenSourceCodeVisitor thenv = mv.visitIf("hasBeenSet");
			/*
			 * Skip property whose value has not been set. Properties may be
			 * either declared in a content class or dynamically defined by
			 * calling setValue() on the property controller. Note that
			 * setValue() may also be called for declared properties. Here we
			 * only want to deal with properties which have been set.
			 */
			thenv.visitSet("Object value","weaveableSCAPC.getValue(\""+name+"\")");
			generateSetInjectionPointValue(mv,ip,"value");
			thenv.visitEnd();
		}

		// Inject POJO defined references and properties
		for (Map.Entry<String,Method> entry : ccmd.setters.entrySet()) {
			String name = entry.getKey();
			Method setter = entry.getValue();
			Class<?>[] ptypes = setter.getParameterTypes();

			CatchSourceCodeVisitor catchv = mv.visitTry();
			// Check whether this is a reference
			catchv.visitSet(ComponentType.class.getName()+" ct","("+ComponentType.class.getName()+") weaveableC.getFcType()");
			catchv.visitSet(InterfaceType.class.getName()+" it","ct.getFcInterfaceType(\""+name+"\")");
			// Should we test that this is a client interface?
			catchv.visitSet("Object sr","getServiceReference(it)");
			catchv.visitIns("content.",setter.getName(),"((",ptypes[0].getName(),")sr)");
			catchv = catchv.visitCatch(NoSuchInterfaceException.class,"nsie");
			// Check whether this is a property
			ThenSourceCodeVisitor thenv = catchv.visitIf("weaveableSCAPC.containsPropertyName(\"",name,"\")");
			thenv.visitSet("Object value","weaveableSCAPC.getValue(\"",name,"\")");
			catchv.visitIns("content.",setter.getName(),"((",ptypes[0].getName(),")value)");
			thenv.visitEnd();
			/*
			 * Else, this is neither a reference, nor a property. We do not
			 * inject anything in this setter.
			 */
			catchv.visitEnd();
		}

		/*
		 * Register all field properties with the property controller.
		 *
		 * The idea is that field-based properties may have a default expression
		 * in the Java code, and that this value should be registered with the
		 * property controller in order for it to be able to return this value
		 * when SCAPropertyController#getValue(String) is invoked.
		 */
		ThenSourceCodeVisitor thenv = null;
		for (Map.Entry<String,InjectionPoint<Annotation>> entry : ccmd.props.entrySet()) {
			String name = entry.getKey();
			InjectionPoint<Annotation> ip = entry.getValue();
			if( ip instanceof InjectionPointFieldImpl ) {
				Class<?> type = ip.getType();
				String tname = stringifyInstantiation(type);
				if( thenv == null ) {
					thenv = mv.visitIf("!instantiatedOnce");
					generateGetInjectionPointValue(thenv,ip,"value","Object");
				}
				else {
					generateGetInjectionPointValue(thenv,ip,"value","");
				}
				thenv.visitIns("weaveableSCAPC.setValue(\""+name+"\",value)");
				thenv.visitIns("weaveableSCAPC.setType(\""+name+"\",",tname,")");
			}
		}
		if( thenv != null ) {
			thenv.visitEnd();
		}

		// Check that all mandatory properties have been set
		for (Map.Entry<String,InjectionPoint<Annotation>> entry : ccmd.props.entrySet()) {
			String name = entry.getKey();
			InjectionPoint<Annotation> ip = entry.getValue();
			boolean isRequired = isRequired(ip);
			if(isRequired) {
				thenv = mv.visitIf("! weaveableSCAPC.hasBeenSet(\""+name+"\")");
				thenv.visitSet("String msg","\"Required property",name,"in component class ",contentClassName," should have been set\"");
				thenv.visitIns("throw new",ContentInstantiationException.class.getName(),"(msg)");
				thenv.visitEnd();
			}
		}

		// Inject callbacks
		if( ccmd.callbacks.length > 0 ) {
			mv.visitSet(CallbackManager.class.getName()+" cm",CallbackManager.class.getName(),".get()");
			thenv = mv.visitIf("! cm.isEmpty()");
			thenv.visitSet(ServiceReference.class.getName()+" cr","cm.peek()");
			thenv.visitSet("Class businessItf","cr.getBusinessInterface()");
			for(InjectionPoint<?> ip : ccmd.callbacks) {
				Class<?> type = ip.getType();
				String tname = stringifyInstantiation(type);
				ThenSourceCodeVisitor then2v = thenv.visitIf(tname,".isAssignableFrom(businessItf)");
				generateSetInjectionPointValue(then2v,ip,"cr");
				then2v.visitEnd();
			}
			thenv.visitEnd();
		}

		// Invoke the @Init annotated method, if any
		if( ccmd.initMethod != null ) {
			String name = ccmd.initMethod.getName();
			mv.visitIns("content.",name,"()");
		}
	}

	// --------------------------------------------------------------------
	// Generation of the getFcCurrentContents method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodGetFcCurrentContents( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PRIVATE, null, contentClassName+"[]",
				"getFcCurrentContents", null, null );
		 generateMethodGetFcCurrentContents(mv);
		 mv.visitEnd();
	}

	/**
	 * @since 1.4.5
	 */
	protected abstract void generateMethodGetFcCurrentContents( BlockSourceCodeVisitor mv );

	// --------------------------------------------------------------------
	// Generation of the setPropertyValue method
	// --------------------------------------------------------------------

	protected void generateMethodSetPropertyValue( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "void",
				"setPropertyValue",
				new String[]{"String name","Object value"}, null );
		if( ccmd.props.size() != 0 ) {
			// If no property, empty method
			generateMethodSetPropertyValue(mv);
		}
		mv.visitEnd();
	}

	/**
	 * @since 1.6
	 */
	protected void generateMethodGetPropertyValue( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "Object",
				"getPropertyValue",
				new String[]{"String name"}, null );
		if( ccmd.props.size() == 0 ) {
			// If no property, return null
			mv.visitIns("return","null");
		}
		else {
			generateMethodGetPropertyValue(mv);
		}
		mv.visitEnd();
	}

	/**
	 * @since 1.6
	 */
	private void generateMethodGetPropertyValue( BlockSourceCodeVisitor mv ) {

		Set<Map.Entry<String,InjectionPoint<Annotation>>> set = ccmd.props.entrySet();
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> props = set.iterator();

		mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
		ThenSourceCodeVisitor thenv = mv.visitIf("contents.length == 0");
		thenv.visitIns("return","null");
		thenv.visitEnd();

		Map.Entry<String,InjectionPoint<Annotation>> entry = props.next();
		InjectionPoint<Annotation> ip = entry.getValue();
		mv.visitSet(contentClassName+" content","contents[0]");
		boolean b = generateGetInjectionPointValue(mv,ip,"value","Object");
		if(b) {
			if( props.hasNext() ) {
				BlockSourceCodeVisitor fv =
					mv.visitFor("int i=1 ; i < contents.length ; i++");
				fv.visitSet("content","contents[i]");
				generateMethodGetPropertyValue(fv,props);
				fv.visitEnd();
			}

			mv.visitIns("return","value");
		}
		else {
			// The value of the property cannot be retrieved
			mv.visitIns("return","null");
		}
	}

	/**
	 * @since 1.6
	 */
	private void generateMethodGetPropertyValue(
		BlockSourceCodeVisitor blockv,
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> ips ) {

		Map.Entry<String,InjectionPoint<Annotation>> entry = ips.next();
		String ipname = entry.getKey();
		InjectionPoint<Annotation> ip = entry.getValue();
		String condition = "name.equals(\""+ipname+"\")";
		ThenSourceCodeVisitor thenv = blockv.visitIf(condition);
		boolean b = generateGetInjectionPointValue(thenv,ip,"v","Object");
		if(b) {
			ThenSourceCodeVisitor thenv2 = thenv.visitIf("v != value");
			thenv2.visitIns("return","null");
			thenv2.visitEnd();

			if( ips.hasNext() ) {
				BlockSourceCodeVisitor elsev = thenv.visitElse();
				generateMethodGetPropertyValue(elsev,ips);
				elsev.visitEnd();
			}
			else {
				thenv.visitEnd();
			}
		}
		else {
			thenv.visitEnd();
		}
	}

	private void generateMethodSetPropertyValue( BlockSourceCodeVisitor mv ) {

		Set<Map.Entry<String,InjectionPoint<Annotation>>> set = ccmd.props.entrySet();
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> props = set.iterator();

		mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
		BlockSourceCodeVisitor fv = mv.visitFor("int i=0 ; i < contents.length ; i++");
		fv.visitSet(contentClassName+" content","contents[i]");
		generateMethodSetPropertyValue(fv,props);
		fv.visitEnd();
	}

	private void generateMethodSetPropertyValue(
		BlockSourceCodeVisitor blockv,
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> ips ) {

		Map.Entry<String,InjectionPoint<Annotation>> entry = ips.next();
		String ipname = entry.getKey();
		InjectionPoint<Annotation> ip = entry.getValue();
		String condition = "name.equals(\""+ipname+"\")";
		ThenSourceCodeVisitor thenv = blockv.visitIf(condition);
		generateSetInjectionPointValue(thenv,ip,"value");

		if( ips.hasNext() ) {
			BlockSourceCodeVisitor elsev = thenv.visitElse();
			generateMethodSetPropertyValue(elsev,ips);
			elsev.visitEnd();
		}
		else {
			thenv.visitEnd();
		}
	}

	private void generateSetInjectionPointValue(
		BlockSourceCodeVisitor blockv, InjectionPoint<?> ip, String value ) {

		Class<?> type = ip.getType();
		type = ClassHelper.box(type);
		String tname = ClassHelper.getName(type);
		String castedvalue = "("+tname+")"+value;

		if( ip instanceof InjectionPointFieldImpl ) {
			InjectionPointFieldImpl<?> ipf = (InjectionPointFieldImpl<?>) ip;
			Field f = ipf.getField();
			String fname = f.getName();
			Class<?> dc = f.getDeclaringClass();
			String name = getTargetTypeName();

			if( MemberHelper.isAccessibleWithStaticInvocation(f,name) ) {
				blockv.visitSet("content."+fname,castedvalue);
			}
			else {
				// Use reflection
				CatchSourceCodeVisitor catchv = blockv.visitTry();
				catchv.visitSet(Field.class.getName()+" field",dc,".class.getDeclaredField(\""+fname+"\")");
				catchv.visitIns("field.setAccessible(true)");
				catchv.visitIns("field.set(content,",value,")");
				catchv = catchv.visitCatch(Exception.class,"e");
				catchv.visitIns("throw new",TinfiRuntimeException.class.getName(),"(e)");
				catchv.visitEnd();
			}
		}
		else {
			InjectionPointMethodImpl<?> ipm = (InjectionPointMethodImpl<?>) ip;
			Method setter = ipm.getSetterMethod();
			String mname = setter.getName();
			Class<?> dc = setter.getDeclaringClass();
			String name = getTargetTypeName();

			if( MemberHelper.isAccessibleWithStaticInvocation(setter,name) ) {
				blockv.visitIns("content.",mname,"(",castedvalue,")");
			}
			else {
				// Use reflection
				CatchSourceCodeVisitor catchv = blockv.visitTry();
				catchv.visitSet(Method.class.getName()+" setter",dc,".class.getDeclaredMethod(\""+mname+"\",",type.getName(),".class)");
				catchv.visitIns("setter.setAccessible(true)");
				catchv.visitIns("setter.invoke(content,",value,")");
				catchv = catchv.visitCatch(Exception.class,"e");
				catchv.visitIns("throw new",TinfiRuntimeException.class.getName(),"(e)");
				catchv.visitEnd();
			}
		}
	}

	/**
	 * @return
	 *         <code>true</code> if the code for getting the value stored in the
	 *         injection point can be retrieved, <code>false</code> otherwise. This
	 *         is the case when this is an injection point based on methods but
	 *         that no getter method is available (only a setter in this case.)
	 */
	private boolean generateGetInjectionPointValue(
		BlockSourceCodeVisitor blockv, InjectionPoint<?> ip,
		String valueLocalFieldName, String valueLocalFieldType ) {

		Class<?> type = ip.getType();
		type = ClassHelper.box(type);
		String tname =
			valueLocalFieldType==null ?
			ClassHelper.getName(type) :
			valueLocalFieldType;

		if( ip instanceof InjectionPointFieldImpl ) {
			InjectionPointFieldImpl<?> ipf = (InjectionPointFieldImpl<?>) ip;
			Field f = ipf.getField();
			String fname = f.getName();
			Class<?> dc = f.getDeclaringClass();
			String name = getTargetTypeName();

			if( MemberHelper.isAccessibleWithStaticInvocation(f,name) ) {
				blockv.visitSet(tname+" "+valueLocalFieldName,"content.",fname);
			}
			else {
				// Use reflection
				blockv.visitSet(tname+"  "+valueLocalFieldName,"null");
				CatchSourceCodeVisitor catchv = blockv.visitTry();
				catchv.visitSet(Field.class.getName()+" field",dc,".class.getDeclaredField(\""+fname+"\")");
				catchv.visitIns("field.setAccessible(true)");
				catchv.visitSet(valueLocalFieldName,"field.get(content)");
				catchv = catchv.visitCatch(Exception.class,"e");
				catchv.visitIns("throw new",TinfiRuntimeException.class.getName(),"(e)");
				catchv.visitEnd();
			}
		}
		else {
			InjectionPointMethodImpl<?> ipm = (InjectionPointMethodImpl<?>) ip;
			Method getter = ipm.getGetterMethod();
			if( getter == null ) {
				// No getter method to retrieve the value of this property
				return false;
			}
			else {
				String mname = getter.getName();
				Class<?> dc = getter.getDeclaringClass();
				String name = getTargetTypeName();

				if( MemberHelper.isAccessibleWithStaticInvocation(getter,name) ) {
					blockv.visitSet(tname+" "+valueLocalFieldName,"content.",mname,"()");
				}
				else {
					// Use reflection
					blockv.visitSet(tname+" "+valueLocalFieldName,"null");
					CatchSourceCodeVisitor catchv = blockv.visitTry();
					catchv.visitSet(Method.class.getName()+" getter",dc,".class.getDeclaredMethod(\""+mname+"\")");
					catchv.visitIns("getter.setAccessible(true)");
					catchv.visitSet(valueLocalFieldName,"getter.invoke(content)");
					catchv = catchv.visitCatch(Exception.class,"e");
					catchv.visitIns("throw new",TinfiRuntimeException.class.getName(),"(e)");
					catchv.visitEnd();
				}
			}
		}

		return true;
	}

	// --------------------------------------------------------------------
	// Generation of the setReferenceValue method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodSetReferenceValue( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "void",
				"setReferenceValue",
				new String[]{
					"String name",
					ServiceReference.class.getName()+"<?> value"},
				null );
		if( ccmd.refs.size() != 0 ) {
			// If no reference, empty method
			generateMethodSetReferenceValue(mv);
		}
		mv.visitEnd();
	}

	private void generateMethodSetReferenceValue( BlockSourceCodeVisitor mv ) {

		Set<Map.Entry<String,InjectionPoint<Annotation>>> set = ccmd.refs.entrySet();
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> refs = set.iterator();

		mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
		BlockSourceCodeVisitor fv = mv.visitFor("int i=0 ; i < contents.length ; i++");
		fv.visitSet(contentClassName+" content","contents[i]");
		generateMethodSetReferenceValue(fv,refs);
		fv.visitEnd();
	}

	private void generateMethodSetReferenceValue(
		BlockSourceCodeVisitor blockv,
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> ips ) {

		Map.Entry<String,InjectionPoint<Annotation>> entry = ips.next();
		String ipname = entry.getKey();
		InjectionPoint<Annotation> ip = entry.getValue();
		Class<?> iptype = ip.getType();

		String condition =
			List.class.isAssignableFrom(iptype) ||
			Vector.class.isAssignableFrom(iptype) ||
			Map.class.isAssignableFrom(iptype) ||
			Hashtable.class.isAssignableFrom(iptype) ?
				"name.startsWith(\""+ipname+"\")" :
				"name.equals(\""+ipname+"\")";
		ThenSourceCodeVisitor thenv = blockv.visitIf(condition);
		generateMethodSetReferenceValue(thenv,ip);

		if( ips.hasNext() ) {
			BlockSourceCodeVisitor elsev = thenv.visitElse();
			generateMethodSetReferenceValue(elsev,ips);
			elsev.visitEnd();
		}
		else {
			thenv.visitEnd();
		}
	}

	private void generateMethodSetReferenceValue(
		BlockSourceCodeVisitor blockv, InjectionPoint<?> ip ) {

		Class<?> iptype = ip.getType();

		if( List.class.isAssignableFrom(iptype) ||
			Vector.class.isAssignableFrom(iptype) ) {

			// SCA style collection reference
			generateGetInjectionPointValue(blockv,ip,"srs",null);

			/*
			 * If a service reference has already been recorded for the same
			 * name, we don't need to add it again. The binding may have
			 * changed but, in this case, this is the delegate field of the
			 * component output interceptor which has been updated. The
			 * service reference object which is injected in the injection
			 * point does not need to be changed. The chain service
			 * reference > component interface > component interceptor is
			 * still valid.
			 */
			BlockSourceCodeVisitor forv = null;
			if( List.class.isAssignableFrom(iptype) ) {
				forv = blockv.visitFor("Object sr : srs");
				forv.visitSet(ServiceReferenceImpl.class.getName()+"<?> sri","(",ServiceReferenceImpl.class.getName(),"<?>) sr");
			}
			else {
				// JavaME-style support for multiple references (Vector-based)
				forv = blockv.visitFor(Enumeration.class.getName()+" e = srs.elements() ; e.hasMoreElements() ; ");
				forv.visitSet(ServiceReferenceImpl.class.getName()+" sri","(",ServiceReferenceImpl.class.getName(),") e.nextElement()");
			}
			forv.visitSet("Object o","sri._getDelegate()");
			forv.visitSet(Interface.class.getName()+" itf","(",Interface.class.getName(),") o");
			forv.visitSet("String itfname","itf.getFcItfName()");
			ThenSourceCodeVisitor thenv = forv.visitIf("itfname.equals(name)");
			thenv.visitIns("return");
			thenv.visitEnd();
			forv.visitEnd();

			if( List.class.isAssignableFrom(iptype) ) {
				blockv.visitIns("srs.add(value)");
			}
			else {
				// JavaME-style support for multiple references (Vector-based)
				blockv.visitIns("srs.addElement(value)");
			}
		}
		else if( Map.class.isAssignableFrom(iptype) ||
				 Hashtable.class.isAssignableFrom(iptype) ) {
			// Fraclet style collection reference
			generateGetInjectionPointValue(blockv,ip,"srs",null);
			blockv.visitIns("srs.put(name,value)");
		}
		else {
			// Singleton reference
			generateSetInjectionPointValue(blockv,ip,"value");
		}
	}

	// --------------------------------------------------------------------
	// Generation of the unsetReferenceValue method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodUnsetReferenceValue( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, "void",
				"unsetReferenceValue", new String[]{"String name"}, null );
		if( ccmd.refs.size() != 0 ) {
			// If no reference, empty method
			generateMethodUnsetReferenceValue(mv);
		}
		mv.visitEnd();
	}

	private void generateMethodUnsetReferenceValue( BlockSourceCodeVisitor mv ) {

		Set<Map.Entry<String,InjectionPoint<Annotation>>> set = ccmd.refs.entrySet();
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> refs = set.iterator();

		mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
		BlockSourceCodeVisitor fv = mv.visitFor("int i=0 ; i < contents.length ; i++");
		fv.visitSet(contentClassName+" content","contents[i]");
		generateMethodUnsetReferenceValue(fv,refs);
		fv.visitEnd();
	}

	private void generateMethodUnsetReferenceValue(
		BlockSourceCodeVisitor blockv,
		Iterator<Map.Entry<String,InjectionPoint<Annotation>>> ips ) {

		Map.Entry<String,InjectionPoint<Annotation>> entry = ips.next();
		String ipname = entry.getKey();
		InjectionPoint<Annotation> ip = entry.getValue();
		Class<?> iptype = ip.getType();

		String condition =
			List.class.isAssignableFrom(iptype) ||
			Vector.class.isAssignableFrom(iptype) ||
			Map.class.isAssignableFrom(iptype) ||
			Hashtable.class.isAssignableFrom(iptype) ?
				"name.startsWith(\""+ipname+"\")" :
				"name.equals(\""+ipname+"\")";
		ThenSourceCodeVisitor thenv = blockv.visitIf(condition);
		generateMethodUnsetReferenceValue(thenv,ip);

		if( ips.hasNext() ) {
			BlockSourceCodeVisitor elsev = thenv.visitElse();
			generateMethodUnsetReferenceValue(elsev,ips);
			elsev.visitEnd();
		}
		else {
			thenv.visitEnd();
		}
	}

	private void generateMethodUnsetReferenceValue(
		BlockSourceCodeVisitor blockv, InjectionPoint<?> ip ) {

		Class<?> iptype = ip.getType();

		if( List.class.isAssignableFrom(iptype) ) {

			// SCA style collection reference
			generateGetInjectionPointValue(blockv,ip,"srs",null);

			/*
			 * If a service reference has already been recorded for the same
			 * name, we don't need to add it again. The binding may have
			 * changed but, in this case, this is the delegate field of the
			 * component output interceptor which has been updated. The
			 * service reference object which is injected in the injection
			 * point does not need to be changed. The chain service
			 * reference > component interface > component interceptor is
			 * still valid.
			 */
			blockv.visitSet("Object toBeRemoved","null");
			BlockSourceCodeVisitor forv = null;
			if( List.class.isAssignableFrom(iptype) ) {
				forv = blockv.visitFor("Object sr : srs");
				forv.visitSet(ServiceReferenceImpl.class.getName()+"<?> sri","(",ServiceReferenceImpl.class.getName(),"<?>) sr");
			}
			else {
				// JavaME-style support for multiple references (Vector-based)
				forv = blockv.visitFor(Enumeration.class.getName()+" e = srs.elements() ; e.hasMoreElements() ; ");
				forv.visitSet(ServiceReferenceImpl.class.getName()+" sri","(",ServiceReferenceImpl.class.getName(),") e.nextElement()");
			}
			forv.visitSet("Object o","sri._getDelegate()");
			forv.visitSet(Interface.class.getName()+" itf","(",Interface.class.getName(),") o");
			forv.visitSet("String itfname","itf.getFcItfName()");
			ThenSourceCodeVisitor thenv = forv.visitIf("itfname.equals(name)");
			thenv.visitSet("toBeRemoved","sr");
			thenv.visitEnd();
			forv.visitEnd();

			if( List.class.isAssignableFrom(iptype) ) {
				blockv.visitIns("srs.remove(toBeRemoved)");
			}
			else {
				// JavaME-style support for multiple references (Vector-based)
				blockv.visitIns("srs.removeElement(toBeRemoved)");
			}
		}
		else if( Map.class.isAssignableFrom(iptype) ||
				 Hashtable.class.isAssignableFrom(iptype) ) {
			// Fraclet style collection reference
			generateGetInjectionPointValue(blockv,ip,"srs",null);
			blockv.visitIns("srs.remove(name)");
		}
		else {
			// Singleton reference
			generateSetInjectionPointValue(blockv,ip,"null");
		}
	}

	// --------------------------------------------------------------------
	// Generation of the eagerInit method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodEagerInit( ClassSourceCodeVisitor cv ) {
		if( ccmd.eagerinit ) {
			BlockSourceCodeVisitor mv =
				cv.visitMethod(
					Modifier.PUBLIC,null,"void","eagerInit",null,
					new String[]{ContentInstantiationException.class.getName()});
			generateMethodEagerInit(mv);
			mv.visitEnd();
		}
	}

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodEagerInit( BlockSourceCodeVisitor mv ) {
		// Overriden in SCAContentControllerCompositeScopeClassGenerator
		mv.visitSet("String msg","\"Eager initialization is meaningless in scopes other than COMPOSITE. Scope is: ",ccmd.scope,".\"");
		mv.visitIns("throw new ",ContentInstantiationException.class.getName(),"(msg)");
	}

	// --------------------------------------------------------------------
	// Generation of the start method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodStart( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(Modifier.PUBLIC,null,"void","start",null,null);
		if( ccmd.startMethod != null ) {
			// If no start method, empty method
			mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
			BlockSourceCodeVisitor fv = mv.visitFor("int i=0 ; i < contents.length ; i++");
			fv.visitSet(contentClassName+" content","contents[i]");
			fv.visitIns("content.",ccmd.startMethod.getName(),"()");
			fv.visitEnd();
		}
		mv.visitEnd();
	}


	// --------------------------------------------------------------------
	// Generation of the stop method
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodStop( ClassSourceCodeVisitor cv ) {

		BlockSourceCodeVisitor mv =
			cv.visitMethod(Modifier.PUBLIC,null,"void","stop",null,null);
		if( ccmd.stopMethod != null ) {
			// If no stop method, empty method
			mv.visitSet(contentClassName+"[] contents","getFcCurrentContents()");
			BlockSourceCodeVisitor fv = mv.visitFor("int i=0 ; i < contents.length ; i++");
			fv.visitSet(contentClassName+" content","contents[i]");
			fv.visitIns("content.",ccmd.stopMethod.getName(),"()");
			fv.visitEnd();
		}
		mv.visitEnd();
	}

	// --------------------------------------------------------------------
	// Generation of the containsPropertyName, getPropertyNames and getPropertyType methods
	// --------------------------------------------------------------------

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodContainsPropertyName( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,"boolean","containsPropertyName",
				new String[]{"String name"},null);
		mv.visitIns("return props.containsKey(name)");
		mv.visitEnd();
	}

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodGetPropertyNames( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,"String[]","getPropertyNames",null,null);
		mv.visitIns("return propnames");
		mv.visitEnd();
	}

	/**
	 * @since 1.4.5
	 */
	protected void generateMethodGetPropertyType( ClassSourceCodeVisitor cv ) {
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC,null,"Class","getPropertyType",
				new String[]{"String name"},null);
		mv.visitIns("return (Class) props.get(name)");
		mv.visitEnd();
	}


	// --------------------------------------------------------------------
	// Helper methods
	// --------------------------------------------------------------------

	/**
	 * Return a stringified representation for the instantiation of the
	 * specified type.
	 *
	 * @since 1.4.5
	 */
	private static String stringifyInstantiation( Class<?> cl ) {
		if( cl.isArray() ) {
			String s = stringifyArrayInstantiation(cl);
			return "new "+s+".getClass()";
		}
		return cl.getName()+".class";
	}

	/**
	 * Return a stringified representation for the instantiation of the
	 * specified array type.
	 *
	 * @since 1.4.5
	 */
	private static String stringifyArrayInstantiation( Class<?> cl ) {
		if( cl.isArray() ) {
			Class<?> element = cl.getComponentType();
			return stringifyArrayInstantiation(element)+"[0]";
		}
		String name = ClassHelper.getName(cl);
		return name;
	}

	/**
	 * Retrieve the names and the types of the @{@link Property} annotated
	 * parameters for the specified constructor. The names ant types are added
	 * to the specified lists.
	 *
	 * @param ctor   the constructor
	 * @param names  the parameter names
	 * @param types  the parameter types
	 *
	 * @since 1.4.5
	 */
	private static void retrieveInitParameters(
		Constructor<?> ctor, List<String> names, List<Class<?>> types ) {

		Annotation[][] paramsAnnots = ctor.getParameterAnnotations();
		for (Annotation[] paramAnnots : paramsAnnots) {
			for (Annotation pannot : paramAnnots) {
				if( pannot instanceof Property ) {
					Property propannot = (Property) pannot;
					String name = propannot.name();
					names.add(name);
				}
			}
		}

		Class<?>[] paramTypes = ctor.getParameterTypes();
		types.addAll(Arrays.asList(paramTypes));
	}

	/**
	 * Return <code>true</code> if the specified injection point is associated
	 * with either the annotation parameter <code>required=true</code> or
	 * <code>contingency=MANDATORY</code>. Both cases correspond to injection
	 * points for required references or properties, either SCA style or Fraclet
	 * style.
	 *
	 * @param ip  the injection point
	 * @since 1.4.5
	 */
	private static boolean isRequired( InjectionPoint<?> ip ) {
		Annotation annot = ip.getAnnotation();
		boolean isRequired =  false;
		try {
			isRequired =  // <Boolean> for compilation with javac
				AnnotationHelper.<Boolean>getAnnotationParamValue(annot,"required");
		}
		catch( NullPointerException npe ) {
			/*
			 * NPE occurs when getAnnotationParamValue returns null.
			 * null cannot be assigned to boolean, hence the NPE.
			 *
			 * Check the Fraclet annotation style with the contingency
			 * parameter.
			 */
			Contingency contingency =
				AnnotationHelper.getAnnotationParamValue(annot,"contingency");
			if( contingency == null ) {
				/*
				 * Shouldn't occur. The annotation is either SCA style with a
				 * required parameter or Fraclet style with a contingency
				 * parameter.
				 */
				final String msg =
					"required or contingency parameter expected on annotation "+
					annot;
				throw new TinfiRuntimeException(msg);
			}
			isRequired = contingency.equals(Contingency.MANDATORY);
		}

		return isRequired;
	}

	/**
	 * Return the properties associated with the current content class.
	 *
	 * @since 1.4.5
	 */
	private Map<String,Class<?>> getProperties() {

		// @Property annotated fields and setters
		Map<String,Class<?>> props = new HashMap<>();
		CompositeInjectionPointHashMap ipm = ccmd.props;
		Set<String> ipnames = ipm.keySet();
		for (String ipname : ipnames) {
			InjectionPoint<?> ip = ipm.get(ipname);
			Class<?> type = ip.getType();
			props.put(ipname,type);
		}

		// @Property annotated parameters of @Constructor annotated constructors
		Constructor<?> ctr = ccmd.constructorAnnotatedElement;
		if( ctr != null ) {
			Class<?>[] ptypes = ctr.getParameterTypes();
			Annotation[][] psannots = ctr.getParameterAnnotations();
			for (int i = 0; i < ptypes.length; i++) {
				// By definition, ptypes.length == psannots.length
				Class<?> ptype = ptypes[i];
				Annotation[] pannots = psannots[i];
				for (Annotation pannot : pannots) {
					if( pannot instanceof Property ) {
						Property prop = (Property) pannot;
						String propname = prop.name();
						props.put(propname,ptype);
					}
				}
			}
		}

		return props;
	}
}
