/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.Type;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.core.proxy.AbstractProxyClassGenerator;
import org.osoa.sca.annotations.EndsConversation;
import org.ow2.frascati.tinfi.TinfiComponentCBInterface;

/**
 * This class generates the source code of component callback interface
 * implementations.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 */
public class CallBackInterfaceClassGenerator
extends AbstractProxyClassGenerator {

	public CallBackInterfaceClassGenerator() {}

	public CallBackInterfaceClassGenerator(
		InterfaceType it, Class<?> cl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,cl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source code generator.
	 */
	public boolean match() {
		return true;
	}

	/**
	 * Return the name of the suffix that is appended to generated class names.
	 */
	public String getClassNameSuffix() {
		return "FcCBItf";
	}

	@Override
	public void setMergeable( boolean mergeable ) {
		if(mergeable) {
			final String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	@Override
	public String getSuperClassName() {
		StringBuilder sb = new StringBuilder();
		sb.append(TinfiComponentCBInterface.class.getName());
		sb.append('<');
		sb.append(it.getFcItfSignature());
		sb.append('>');
		return sb.toString();
	}

	@Override
	public String[] getImplementedInterfaceNames() {
		return new String[]{it.getFcItfSignature()};
	}

	@Override
	public void generateConstructors( ClassSourceCodeVisitor cv ) {

		super.generateConstructors(cv);

		// public constructor()
		BlockSourceCodeVisitor mv =
			cv.visitConstructor(Modifier.PUBLIC,null,null,null);
		mv.visitEnd();

		// public constructor(Component,String,Type,boolean,Object)
		String[] parameters =
			new String[]{
				Component.class.getName()+" component",
				"String s",
				Type.class.getName()+" type",
				"boolean flag",
				"Object obj"
			};
		mv = cv.visitConstructor(Modifier.PUBLIC,null,parameters,null);
		mv.visitIns("super(component,s,type,flag,obj);");
		mv.visitEnd();
	}

	@Override
	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, Method proxym ) {

		/*
		 * Register the request context and get the content instance that
		 * corresponds to the current scope policy.
		 */
		mv.visit  ("    ");
		mv.visit  (it.getFcItfSignature());
		mv.visit  (" impl = setRequestContextAndGet(\"");
		mv.visit  (it.getFcItfName());
		mv.visitln("\","+it.getFcItfSignature()+".class,this);");

		// Begin of the try block
		mv.visitln("    try {");

		// Call to the delegate
		generateProxyMethodBodyDelegatingCode(mv,proxym);

		// Return the result
		generateProxyMethodBodyReturn(mv,proxym);

		// End of the try block
		mv.visitln("    }");
		mv.visitln("    finally {");

		// Release the content instance
		mv.visit("      releaseContent(impl,");
		if( proxym.getAnnotation(EndsConversation.class) == null ) {
			mv.visit("false");
		}
		else {
			mv.visit("true");
		}
		mv.visitln(");");

		// End of the finally block
		mv.visitln("    }");
	}

	public String getDelegatingInstance( Method proxym ) {
		return "impl";
	}
}
