/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Philippe Merle
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.lang.reflect.Method;

import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.lang.ClassHelper;
import org.objectweb.fractal.juliac.core.proxy.AbstractInterceptorSourceCodeGenerator;
import org.osoa.sca.annotations.EndsConversation;

/**
 * This class generates the source code of Tinfi interceptors that implement
 * the logic related to the management of component content instances. This
 * interceptor invokes the SCA content controller.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 */
public class SCAContentInterceptorSourceCodeGenerator
extends AbstractInterceptorSourceCodeGenerator {

	public SCAContentInterceptorSourceCodeGenerator() {
		super();
	}

	public SCAContentInterceptorSourceCodeGenerator(
		InterfaceType it, Class<?> cl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,cl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return the name of the suffix which is appended to generated class names.
	 */
	public String getClassNameSuffix() {
		return "SCACC";
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source code generator.
	 */
	public boolean match() {

		// Skip control interfaces
		String itname = it.getFcItfName();
		if( itname.equals("component") || itname.endsWith("-controller") ) {
			return false;
		}

		// Skip client interfaces
		if( it.isFcClientItf() ) {
			return false;
		}

		// Match all other interfaces (business server)
		return true;
	}

	@Override
	public void setMergeable( boolean mergeable ) {
		if(mergeable) {
			final String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	@Override
	public void generateProxyMethodBodyBeforeCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		/*
		 * Regiqter the request context and get the content instance which
		 * corresponds to the current scope policy.
		 */
		mv.visit  ("    ");
		mv.visit  (it.getFcItfSignature());
		String[] tpnames = ClassHelper.getTypeParameterNames(proxycl);
		if( tpnames.length != 0 ) {
			String s = ClassHelper.getTypeParameterNamesSignature(tpnames);
			mv.visit(s);
		}
		mv.visit  (" impl = setRequestContextAndGet(\"");
		mv.visit  (it.getFcItfName());
		mv.visitln("\","+it.getFcItfSignature()+".class,this);");

		// TODO the request context created by setRequestContextAndGet should be given a client proxy (not this which is a server one)

		// Begin of the try block
		mv.visitln("    try {");
	}

	@Override
	public void generateProxyMethodBodyAfterReturningCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		// End of the try block
		mv.visitln("    }");
		mv.visitln("    finally {");

		// Release the content instance
		mv.visit("      releaseContent(impl,");
		if( proxym.getAnnotation(EndsConversation.class) == null ) {
			mv.visit("false");
		}
		else {
			mv.visit("true");
		}
		mv.visitln(");");

		// End of the finally block
		mv.visitln("    }");
	}
}
