/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2008-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributor: Philippe Merle
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.objectweb.fractal.api.Interface;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.lang.ClassHelper;
import org.objectweb.fractal.juliac.commons.lang.reflect.MethodHelper;
import org.objectweb.fractal.juliac.core.proxy.AbstractInterceptorSourceCodeGenerator;
import org.ow2.frascati.tinfi.IntentJoinPointImpl;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.api.IntentHandler;

/**
 * This class generates the source code of Tinfi interceptors that implement
 * the logic related to the management of intents associated with components.
 * This interceptor invokes the SCA intent controller.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @author Philippe Merle <Philippe.Merle@inria.fr>
 * @since 0.3
 */
public class SCAIntentInterceptorSourceCodeGenerator
extends AbstractInterceptorSourceCodeGenerator {

	public SCAIntentInterceptorSourceCodeGenerator() {
		super();
	}

	public SCAIntentInterceptorSourceCodeGenerator(
		InterfaceType it, Class<?> cl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,cl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return the name of the suffix which is appended to generated class names.
	 */
	public String getClassNameSuffix() {
		return "Intent";
	}

	/**
	 * Return <code>true</code> if some code must be generated by the current
	 * source code generator.
	 */
	public boolean match() {
		// Match all interfaces, including control ones
		return true;
	}

	@Override
	public void setMergeable( boolean mergeable ) {
		if(mergeable) {
			final String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	@Override
	public void generateStaticParts( ClassSourceCodeVisitor cv ) {

		/*
		 * Initialize the METHODS field (declared below in method
		 * generateFields).
		 */
		Method[] proxymethods = proxycl.getMethods();
		BlockSourceCodeVisitor bv = cv.visitStaticPart();
		if( proxymethods.length > 0 ) {
			bv.visitln("    try {");
		}
		bv.visitln("      METHODS = new java.lang.reflect.Method[]{");

		// Iterate on methods
		String proxyclname = proxygcl.toString(proxycl);
		for (int i = 0; i < proxymethods.length; i++) {

			Method proxym = proxymethods[i];
			String proxymname = proxym.getName();
			Class<?>[] paramtypes = proxym.getParameterTypes();

			bv.visit("        ");
			bv.visit(proxyclname);
			bv.visit(".class.getMethod(\"");
			bv.visit(proxymname);
			bv.visit("\"");
			for (Class<?> paramtype : paramtypes) {
				String paramtypename = ClassHelper.getName(paramtype);
				bv.visit(",");
				bv.visit(paramtypename);
				bv.visit(".class");
			}
			bv.visitln("),");
		}
		bv.visitln("      };");

		if( proxymethods.length > 0 ) {
			bv.visitln("    }");
			bv.visitln("    catch( NoSuchMethodException e ) {");
			bv.visit  ("      throw new ");
			bv.visit  (TinfiRuntimeException.class.getName());
			bv.visitln("(e);");
			bv.visitln("    }");
		}
		bv.visitEnd();
	}

	@Override
	public void generateFields( ClassSourceCodeVisitor cv ) {

		/*
		 * Generate a METHODS field to store the array of component methods
		 * generated by the interceptor.
		 */
		int modifiers = Modifier.PRIVATE + Modifier.STATIC;
		String type = "java.lang.reflect.Method[]";
		String name = "METHODS";
		cv.visitField( modifiers, type, name, null );

		/*
		 * The following piece of code corresponds to a feature which has been
		 * removed in version 1.3.
		 *
		 * Generating a subclass of IntentJoinPointImpl per method defined in
		 * proxycl enables invoking statically the base method which corresponds
		 * to the intercepted join point. Following a discussion with Philippe,
		 * it has been decided that we wouldn't loose much by performing a
		 * dynamic method invocation instead. The advantage is that we don't
		 * need these subclasses anymore. I keep this piece of code commented
		 * for the record just in case we would need to re-install it in the
		 * future.
		 *
		 * Two other pieces of code related to this feature have been commented:
		 * see IntentJoinPointImpl#proceed() and
		 * see #generateProxyMethodBodyBeforeCode().
		 */
//        /*
//         * Generate a subclass of IntentJoinPointImpl per method defined in
//         * proxycl. Instances of the generated class reifies invocations on the
//         * associated method.
//         */
//        UnifiedMethod[] proxymethods = proxycl.getMethods();
//        String proxyclname = proxycl.getName();
//        for (int i = 0; i <proxymethods.length; i++) {
//
//            UnifiedMethod proxym = proxymethods[i];
//            String proxymname = proxym.getName();
//            String proxymrettypename = proxym.getReturnType().getName();
//            UnifiedClass[] paramtypes = proxym.getParameterTypes();
//
//            // Inner class header
//            ClassSourceCodeVisitor innercv = cv.visitInnerClass();
//            int innermodifiers = Modifier.PRIVATE + Modifier.STATIC;
//            String innername = "IntentJoinPointImplForMethod"+i;
//            innercv.visit(
//                innermodifiers, innername, null,
//                IntentJoinPointImpl.class.getName()+'<'+proxyclname+'>', null );
//
//            // proceed() method
//            MethodSourceCodeVisitor mv =
//                innercv.visitMethod(
//                    Modifier.PUBLIC, null, "Object", "proceed", null,
//                    new String[]{"Throwable"});
//            mv.visitBegin();
//            mv.visitln("    if( index == handlers.size() ) {");
//            mv.visit  ("      ");
//            if( ! proxymrettypename.equals("void") ) {
//                mv.visit("return ");
//            }
//            mv.visit("intentTarget.");
//            mv.visit(proxymname);
//            mv.visit("(");
//            for (int j = 0; j < paramtypes.length; j++) {
//                UnifiedClass paramtype = paramtypes[j];
//                if( paramtypes[j].isPrimitive() ) {
//                    paramtype = paramtype.box();
//                }
//                String paramtypename = paramtype.getName();
//                if(j!=0) mv.visit(",");
//                if( ! paramtypename.equals("java.lang.Object") ) {
//                    mv.visit("(");
//                    mv.visit(paramtypename);
//                    mv.visit(")");
//                }
//                mv.visit("intentMethodArguments[");
//                mv.visit(Integer.toString(j));
//                mv.visit("]");
//            }
//            mv.visitln(");");
//            if( proxymrettypename.equals("void") ) {
//                mv.visitln("      return null;");
//            }
//            mv.visitln("    }");
//            mv.visitln("    else {");
//            mv.visitln("      return super.proceed();");
//            mv.visitln("    }");
//
//            // End of the proceed() method
//            mv.visitEnd();
//
//            /*
//             * Up to r1958 I was generating a base() method for the reason
//             * explained in the next paragraph. See
//             * http://svn.forge.objectweb.org/cgi-bin/viewcvs.cgi/frascati/trunk/tinfi/opt/oo/src/main/java/org/ow2/frascati/tinfi/opt/oo/SCAIntentInterceptorSourceCodeGenerator.java?rev=1958&view=auto
//             * The issue of the two cast(Object) was indirectly resolved by
//             * Stephane Drapeau when I mentioned it: he noticed that this was an
//             * error and removed the inadequate one. It is not clear to me
//             * whether I should expect the issue to reappear with another
//             * interface. I guess not and decide to include the code from the
//             * base method into proceed.
//             *
//             * In theory, the base() method is not strictly needed. One should
//             * be able to invoke the base method from the proceed method
//             * generated just before. It happens that the JDT compiler cannot
//             * decide which method to invoke in the case of the ComponentContext
//             * interface with the two cast(Object) methods. The fact of
//             * introducing the base() method solves the issue.
//             */
//
//            // End of the inner class
//            innercv.visitEnd();
//        }
	}

	@Override
	public void generateMethodInitFcController( BlockSourceCodeVisitor mv ) {
		mv.visitln("    initIntentHandlersMap(METHODS);");
	}

	@Override
	public void generateMethodClone( BlockSourceCodeVisitor mv ) {
		mv.visitln("    clone.initIntentHandlersMap(METHODS);");
	}

	@Override
	public void generateProxyMethodBodyBeforeCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		/*
		 * Retrieve the index of proxym in the array of methods for proxycl.
		 */
		int idx = -1;
		String proxymstring = proxym.toString();
		Method[] ums = proxycl.getMethods();
		for (int i = 0; i < ums.length; i++) {
			Method um = ums[i];
			String umstring = um.toString();
			if( umstring.equals(proxymstring) ) {
				idx = i;
				break;
			}
		}
		if( idx == -1 ) {
			final String msg =
				proxym+" should have been found in the array of methods of "+
				proxycl;
			throw new TinfiRuntimeException(msg);
		}

		// Retrieve the handlers associated with the component
		mv.visit  ("    java.util.List<");
		mv.visit  (IntentHandler.class.getName());
		mv.visit  ("> handlers = intentHandlersMap.get(METHODS[");
		mv.visit  (Integer.toString(idx));
		mv.visitln("]);");

		// Begin of try/finally block
		mv.visitln("    try {");

		/*
		 * If the handlers list is empty, don't create the IntentJoinPoint
		 * instance and directly call the delegate.
		 */
		String proxymname = proxym.getName();
		Class<?>[] ptypes = proxym.getParameterTypes();
		String rtypename = proxygcl.getGenericReturnType(proxym);
		String[] tpnames = MethodHelper.getTypeParameterNames(proxym);

		mv.visitln("      if( handlers.size() == 0 ) {");

		// base call
		mv.visit("        ");
		if( ! rtypename.equals("void") ) {
			mv.visit(rtypename+" ret = ");
		}

		mv.visit("impl.");
		if( tpnames.length > 0 ) {
			mv.visit("<");
			for (int i = 0; i < tpnames.length; i++) {
				if(i>0) { mv.visit(","); }
				mv.visit(tpnames[i]);
			}
			mv.visit(">");
		}
		mv.visit(proxymname);
		mv.visit("(");
		for (int j = 0; j < ptypes.length; j++) {
			if(j!=0) { mv.visit(","); }
			mv.visit("arg");
			mv.visit(Integer.toString(j));
		}
		mv.visitln(");");
		if( ! rtypename.equals("void") ) {
			mv.visitln("        return ret;");
		}
		mv.visitln("      }");

		/*
		 * Some intent handlers are defined.
		 */
		mv.visitln("      else {");

		// Retrieve the interface associated with the interceptor
		mv.visit  ("        ");
		mv.visit  (Interface.class.getName());
		mv.visitln(" itf = getFcItf();");

		// Intent join point creation and initialization
		mv.visit  ("        ");
		mv.visit  (IntentJoinPointImpl.class.getName());
		mv.visit  ("<");
		mv.visit  (it.getFcItfSignature());
		mv.visit  ("> ijp = new ");

		mv.visit  (IntentJoinPointImpl.class.getName());

		/*
		 * The following commented lines of code correspond to a feature
		 * which has been removed in version 1.3.
		 *
		 * See #generateFields() for further information.
		 */
//        mv.visit  ("IntentJoinPointImplForMethod");
//        mv.visit  (Integer.toString(idx));

		mv.visitln("();");
		mv.visit  ("        ijp.init(handlers,fcComp,itf,impl,METHODS[");
		mv.visit  (Integer.toString(idx));
		mv.visit  ("]");
		for (int i = 0; i < ptypes.length; i++) {
			/*
			 * The cast to (Object) is needed to accomodate varags parameters.
			 * The fix was suggested in version 1.4.2 by Philippe.
			 */
			mv.visit(",(Object)arg");
			mv.visit(Integer.valueOf(i).toString());
		}
		mv.visitln(");");
	}

	@Override
	public void generateProxyMethodBodyAfterReturningCode(
		BlockSourceCodeVisitor mv, Method proxym ) {

		// End of else block
		mv.visitln("      }");

		// End of try block
		mv.visitln("    }");
		mv.visitln("    catch( Throwable t ) {");

		// Handle exceptions thrown by the method
		Class<?>[] excepts = proxym.getExceptionTypes();
		for (Class<?> except : excepts) {
			mv.visit  ("      if( t instanceof ");
			mv.visit  (except.getName());
			mv.visitln(" ) {");
			mv.visit  ("        throw (");
			mv.visit  (except.getName());
			mv.visitln(") t;");
			mv.visitln("      }");
		}

		// Handle RuntimeException
		mv.visitln("      if( t instanceof RuntimeException ) {");
		mv.visitln("        throw (RuntimeException) t;");
		mv.visitln("      }");

		// Handle other types of exception
		mv.visit  ("      throw new ");
		mv.visit  (TinfiRuntimeException.class.getName());
		mv.visitln("(t);");

		// End of the catch block
		mv.visitln("    }");
	}
}
