/***
 * OW2 FraSCAti Tinfi
 * Copyright (C) 2007-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: frascati@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.ow2.frascati.tinfi.opt.oo;

import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;

import org.oasisopen.sca.annotation.Callback;
import org.oasisopen.sca.annotation.OneWay;
import org.objectweb.fractal.api.Component;
import org.objectweb.fractal.api.type.InterfaceType;
import org.objectweb.fractal.julia.ComponentInterface;
import org.objectweb.fractal.julia.type.BasicInterfaceType;
import org.objectweb.fractal.juliac.api.desc.MembraneDesc;
import org.objectweb.fractal.juliac.api.generator.SourceCodeGeneratorItf;
import org.objectweb.fractal.juliac.api.visit.BlockSourceCodeVisitor;
import org.objectweb.fractal.juliac.api.visit.ClassSourceCodeVisitor;
import org.objectweb.fractal.juliac.commons.lang.ClassHelper;
import org.objectweb.fractal.juliac.commons.lang.annotation.AnnotationHelper;
import org.objectweb.fractal.juliac.commons.lang.reflect.AnnotatedElementHelper;
import org.objectweb.fractal.juliac.core.helper.InterfaceTypeHelper;
import org.objectweb.fractal.juliac.core.proxy.AbstractProxyClassGenerator;
import org.osoa.sca.annotations.Conversational;
import org.osoa.sca.annotations.EndsConversation;
import org.ow2.frascati.tinfi.TinfiRuntimeException;
import org.ow2.frascati.tinfi.oasis.ServiceReferenceImpl;
import org.ow2.frascati.tinfi.osoa.ConversationImpl;

/**
 * This class generates the source code of SCA service reference
 * implementations.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 */
public class ServiceReferenceClassGenerator
extends AbstractProxyClassGenerator {

	public ServiceReferenceClassGenerator() {}

	public ServiceReferenceClassGenerator(
		InterfaceType it, Class<?> cl, String pkgRoot,
		MembraneDesc<?> membraneDesc, boolean mergeable ) {

		super(it,cl,pkgRoot,membraneDesc,mergeable);
	}

	/**
	 * Return <code>true</code> if some code must be generated by this
	 * generator.
	 */
	public boolean match() {
		return true;
	}

	/**
	 * Return the name of the suffix which is appended to generated class names.
	 */
	public String getClassNameSuffix() {
		return "FcSR";
	}

	@Override
	public void setMergeable( boolean mergeable ) {
		if(mergeable) {
			final String msg = "This generator cannot generate mergeable code";
			throw new IllegalArgumentException(msg);
		}
	}

	@Override
	public String getSuperClassName() {

		StringBuilder sb = new StringBuilder();
		sb.append(ServiceReferenceImpl.class.getName());
		sb.append('<');

		/*
		 * Add type parameters if any.
		 */
		String signature = it.getFcItfSignature();
		sb.append(signature);
		String[] tpnames = ClassHelper.getTypeParameterNames(proxycl);
		if( tpnames.length != 0 ) {
			String s = ClassHelper.getTypeParameterNamesSignature(tpnames);
			sb.append(s);
		}

		sb.append('>');
		return sb.toString();
	}

	@Override
	public String[] getImplementedInterfaceNames() {

		/*
		 * Create the signature of the proxy interface with generic parameters
		 * if any.
		 */
		String signature = it.getFcItfSignature();
		String[] tpnames = ClassHelper.getTypeParameterNames(proxycl);
		if( tpnames.length != 0 ) {
			String s = ClassHelper.getTypeParameterNamesSignature(tpnames);
			signature += s;
		}

		return new String[]{signature};
	}

	@Override
	public void generateConstructors( ClassSourceCodeVisitor cv ) {

		super.generateConstructors(cv);

		/*
		 * Create the signature of the proxy interface with generic parameters
		 * if any.
		 */
		String signature = it.getFcItfSignature();
		String[] tpnames = ClassHelper.getTypeParameterNames(proxycl);
		if( tpnames.length != 0 ) {
			String s = ClassHelper.getTypeParameterNamesSignature(tpnames);
			signature += s;
		}

		// public constructor(Class<B>,B)
		String[] parameters =
			new String[]{
				"Class<"+signature+"> businessInterface",
				signature+" service" };
		BlockSourceCodeVisitor mv =
			cv.visitConstructor(Modifier.PUBLIC,null,parameters,null);
		mv.visitIns("super(businessInterface,service)");
		mv.visitEnd();
	}

	/**
	 * @since 1.4.1
	 */
	@Override
	public void generateMethods( ClassSourceCodeVisitor cv ) {

		super.generateMethods(cv);

		/*
		 * Override the getService method inherited from ServiceReferenceImpl.
		 * The idea is that in generated subclasses of ServiceReferenceImpl we
		 * can directly return "this" for getService. Whereas when getService is
		 * invoked on an instance of ServiceReferenceImpl, getService should
		 * return an instance which implements both the ServiceReference
		 * interface and the business interface. This is the case of generated
		 * subclass instances.
		 */
		String signature = it.getFcItfSignature();
		BlockSourceCodeVisitor mv =
			cv.visitMethod(
				Modifier.PUBLIC, null, signature, "getService", null, null );
		mv.visitIns("return this");
		mv.visitEnd();
	}

	private Conversational conversationalItf;
	private Annotation oneWayMeth;
	private Class<?> callbackClass;

	@Override
	public void generateProxyMethodBody(
		BlockSourceCodeVisitor mv, Method proxym ) {

		conversationalItf = proxycl.getAnnotation(Conversational.class);
		oneWayMeth =
			AnnotatedElementHelper.getAnnotation(
				proxym,
				OneWay.class.getName(),
				"org.osoa.sca.annotations.OneWay");

		callbackClass = Void.class;   // callback.value() defaults to Void
		Annotation callback =
			ClassHelper.getAnnotation(
				proxycl,
				Callback.class.getName(),
				"org.osoa.sca.annotations.Callback");
		if( callback != null ) {
			callbackClass = AnnotationHelper.getAnnotationParamValue(callback,"value");
		}

		// OneWay method
		if( oneWayMeth != null ) {
			Class<?> rtype = proxym.getReturnType();
			if( ! (rtype.isAssignableFrom(void.class)) ) {
				final String msg =
					"@OneWay annotated method "+proxym.toString()+" in "+
					proxygcl.toString(proxycl)+" should return void";
				throw new TinfiRuntimeException(msg);
			}
			mv.visitln("    new Thread() {");
			mv.visitln("      public void run() {");
		}

		// Method defined in a conversational interface
		if( conversationalItf != null ) {
			mv.visitln("        "+ConversationImpl.class.getName()+" conv = _getConversation();");
			mv.visitln("        _pushConversation(conv);");
		}

		// Method defined in an interface with a callback
		if( callbackClass != Void.class ) {

			InterfaceType cbit =
				new BasicInterfaceType(
					"callback",callbackClass.getName(),false,false,false);
			SourceCodeGeneratorItf cg =
				new CallBackInterfaceClassGenerator(
					cbit,callbackClass,pkgRoot,null,mergeable);
			String cbitfclname = cg.getTargetTypeName();

			cg =
				new ServiceReferenceClassGenerator(
					cbit,callbackClass,pkgRoot,null,mergeable);
			String cbcrclname = cg.getTargetTypeName();

			mv.visit  ("        "+ComponentInterface.class.getName()+" ci = (");
			mv.visitln(ComponentInterface.class.getName()+") service;");
			mv.visitln("        "+Component.class.getName()+" owner = ci.getFcItfOwner();");
			mv.visit  ("        "+InterfaceType.class.getName()+" it = ");
			mv.visit  (InterfaceTypeHelper.javaify(cbit).toString());
			mv.visitln(";");
			mv.visitln("        "+cbitfclname+" cbci = new "+cbitfclname+"(owner,\"callback\",it,false,null);");

			mv.visit  ("        "+ServiceReferenceImpl.class.getName()+"<"+callbackClass.getName()+"> cb = ");
			mv.visit  ("new "+cbcrclname+"(");
			mv.visitln(callbackClass.getName()+".class,cbci);");
			mv.visitln("        _setCallback(cb);");
			mv.visitln("        _pushCallback(cb);");
		}

		// Call to the delegate
		generateProxyMethodBodyDelegatingCode(mv,proxym);

		EndsConversation end = proxym.getAnnotation(EndsConversation.class);
		if( end != null ) {
			if( conversationalItf == null ) {
				/*
				 * For cases where @EndsConversation appears in non
				 * @Conversation annotated interfaces. That can be the case for
				 * example for callback interfaces associated with conversation
				 * scoped components.
				 */
				mv.visitln("    "+ConversationImpl.class.getName()+" conv = _peekConversation();");
			}
			mv.visitln("    conv.end();");
		}
		if( conversationalItf != null ) {
			mv.visitln("    conv.access();");
			mv.visitln("    _popConversation();");
		}
		if( callbackClass != Void.class ) {
			mv.visitln("        _popCallback();");
		}

		// Return the result
		generateProxyMethodBodyReturn(mv,proxym);

		if( oneWayMeth != null ) {
			mv.visitln("      }");
			mv.visitln("    }.start();");
		}
	}

	public String getDelegatingInstance( Method proxym ) {
		return "service";
	}
}
