/***
 * Juliac
 * Copyright (C) 2008-2021 Inria, Univ. Lille
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 * Contributors: Valerio Schiavoni, Philippe Merle
 */

package org.objectweb.fractal.juliac.compile.jdt;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.jdt.internal.compiler.ClassFile;
import org.eclipse.jdt.internal.compiler.Compiler;
import org.eclipse.jdt.internal.compiler.IErrorHandlingPolicy;
import org.eclipse.jdt.internal.compiler.classfmt.ClassFileConstants;
import org.eclipse.jdt.internal.compiler.env.ICompilationUnit;
import org.eclipse.jdt.internal.compiler.env.INameEnvironment;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.problem.DefaultProblemFactory;
import org.objectweb.fractal.juliac.api.JuliacItf;
import org.objectweb.fractal.juliac.api.SourceFileItf;

/**
 * This class extends the JDT compiler to provide a reporting mechanism that
 * relates errors to source files and that enables to retrieve compiled files.
 *
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille.fr>
 * @since 2.2.1
 */
public class JDTCompiler extends Compiler {

	/**
	 * @since 2.6
	 */
	public JDTCompiler( JuliacItf jc, List<SourceFileItf> sfs ) {

		super(
			getNameEnvironment(jc,sfs),
			getErrorHandlingPolicy(),
			getCompilerOptions(jc),
			new CompilerRequestorImpl(jc),
			new DefaultProblemFactory() );
	}

	public void compile( Map<ICompilationUnit,SourceFileItf> units ) {

		// Initialize the callback object which receives compilation results
		CompilerRequestorImpl r = (CompilerRequestorImpl) requestor;
		r.init(units);

		// Compile
		Set<ICompilationUnit> luts = units.keySet();
		ICompilationUnit[] uts =
			luts.toArray(new ICompilationUnit[luts.size()]);
		compile(uts);
	}

	public List<ClassFile> getClassFiles() {
		CompilerRequestorImpl r = (CompilerRequestorImpl) requestor;
		List<ClassFile> classFiles = r.getClassFiles();
		return classFiles;
	}

	/**
	 * @since 2.6
	 */
	private static CompilerOptions getCompilerOptions( JuliacItf jc ) {

		CompilerOptions options = new CompilerOptions();

		options.complianceLevel =
			CompilerOptions.releaseToJDKLevel(jc.getJavaReleaseVersionNumber());
		options.originalComplianceLevel = options.complianceLevel;
		options.sourceLevel = options.complianceLevel;
		options.originalSourceLevel = options.complianceLevel;
		options.targetJDK = options.complianceLevel;

		/*
		 * Contribution by Loris.
		 *
		 * Generate debug info for debugging the output.
		 * Tricks like "boolean stopHere = true;" depend on this setting.
		 *
		 * TODO: This should be transmitted as parameter for example with -X
		 */
		options.produceDebugAttributes =
			ClassFileConstants.ATTR_VARS |
			ClassFileConstants.ATTR_LINES |
			ClassFileConstants.ATTR_SOURCE;
		options.preserveAllLocalVariables = true;

		return options;
	}

	/**
	 * @since 2.6
	 */
	private static INameEnvironment getNameEnvironment(
		JuliacItf jc, List<SourceFileItf> sfs ) {

		ClassLoader classLoader = jc.getClassLoader();

		List<String> sources = new ArrayList<>(sfs.size());
		for (SourceFileItf sf : sfs) {
			String qname = sf.getQname();
			sources.add(qname);
		}

		INameEnvironment environment =
			new NameEnvironmentImpl(classLoader,sources);

		return environment;
	}

	private static IErrorHandlingPolicy getErrorHandlingPolicy() {
		return new IErrorHandlingPolicy() {
			public boolean proceedOnErrors() { return true; }
			public boolean stopOnFirstError() { return false; }
			public boolean ignoreAllErrors() { return false; }
		};
	}
}
